# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Registration of messages.

Bind additional information about a message to its class definition. This
includes more complete type information about its members but also settings
associated with how an instance of the message should be treated, e.g. during
logging.

.. autoclass:: Message
.. autoclass:: Unknown
.. autoclass:: Incognito

.. autofunction:: default_clock
.. autofunction:: default_span
.. autofunction:: default_world
.. autofunction:: default_delta
.. autofunction:: default_uuid
.. autofunction:: default_array
.. autofunction:: default_vector
.. autofunction:: default_set
.. autofunction:: default_map
.. autofunction:: default_deque
.. autofunction:: default_none

.. autofunction:: bind_message

.. autofunction:: is_message
.. autofunction:: is_message_class
.. autofunction:: equal_to
"""

__docformat__ = 'restructuredtext'

import typing
import uuid
from datetime import MINYEAR, datetime, timedelta
from enum import Enum
from .virtual_memory import *
from .convert_memory import *
from .convert_signature import *
from .convert_type import *
from .virtual_runtime import *

__all__ = [
	'MessageError',
	'MessageRegistrationError',
	'TypeTrack',
	'correct_track',

	'Message',
	'Unknown',
	'Incognito',

	'DEFAULT_STRING',
	'DEFAULT_UNICODE',
	'DEFAULT_CLOCK',
	'DEFAULT_SPAN',
	'DEFAULT_WORLD',
	'DEFAULT_DELTA',
	'DEFAULT_ZONE',
	'DEFAULT_UUID',

	# No parameters.
	'default_byte',
	'default_character',
	'default_rune',
	'default_block',
	'default_string',
	'default_unicode',
	'default_clock',
	'default_span',
	'default_world',
	'default_delta',
	'default_uuid',
	'default_vector',
	'default_set',
	'default_map',
	'default_deque',
	'default_none',

	# Require parameters.
	'default_array',

	'fix_expression',
	'is_message',
	'is_message_class',
	'compile_schema',

	'bind_message',
	'equal_to',
]

#
class MessageError(Exception):
	"""Base exception for all message exceptions."""

class MessageRegistrationError(MessageError):
	"""A request to register a class cannot be fulfilled.

	:param name: the name of the class being registered
	:param reason: a short description
	"""

	def __init__(self, name: str, reason: str):
		"""Refer to class."""
		self.name = name
		self.reason = reason
		super().__init__(reason)

	def __str__(self):
		"""Compose a readable diagnostic."""
		if self.name:
			return f'cannot register "{self.name}" ({self.reason})'
		return f'registration failed ({self.reason})'

# Holds nested names that would be helpful in the event
# of an error.
class TypeTrack(Exception):
	"""Construct a readable name for a message.member.member, during exceptions."""

	def __init__(self, name, reason):
		"""Keep building the trace."""
		super().__init__(reason)
		if name:
			self.path = [name]
		else:
			self.path = []
		self.reason = reason

def correct_track(e):
	"""Generate a readable version of a TypeTrack exception."""
	t = '.'.join(reversed(e.path))
	return t

#
#
class Message(object):
	"""Internal placeholder class used for dispatching."""

class Unknown(object):
	"""An abstract class used to indicate an unexpected message."""

class Incognito(object):
	"""A class that holds the recovered materials of an unregistered message.

	:param type_name: portable identity of the associated word
	:param decoded_word: parsed but unmarshaled object
	:param saved_pointers: table of :class:`~.Pointer` values
	:type saved_pointers: internal
	:param address_book: table of :class:`~.Address` values
	:type address_book: internal
	"""

	def __init__(self, type_name: str=None, decoded_word: Word=None, saved_pointers=None, address_book=None):
		"""Refer to class."""
		self.type_name = type_name
		self.decoded_word = decoded_word
		self.saved_pointers = saved_pointers
		self.address_book = address_book or {}

# A group of functions that exist to allow type descriptions for
# messages to make use of *classes* rather than *instances* of those
# classes;
#	 VectorOf(Float8)
# instead of;
#	 VectorOf(Float8())
# Unclear on best design/engineering response to the issue but
# certainly this results in fewer parentheses and quicker development.
# Fully correct declaration of user-defined messages quite verbose
# and consequently less clear;
#	 VectorOf(SomeMessage)
# vs;
#	 VectorOf(UserDefined(SomeMessage))

# Default initializers (i.e. no parameters available) for tricky
# types. Want an instance of the type but preferrably without
# side-effects and at the least cost. Most containers need
# a fresh instance of themselves (i.e. list,
# deque...) but other types are objects that are immutable,
# such as bytes and datetime. These can be initialized with a
# single constant value to reduce the cycles consumed by
# initialization. This whole issue is further complicated by
# the merging of the two type systems - kipjak and python - and
# how String does not map to str.

# Immutable initializer constants.
DEFAULT_STRING = bytes()
DEFAULT_UNICODE = str()
DEFAULT_CLOCK = float(0)
DEFAULT_SPAN = float(0)
DEFAULT_WORLD = datetime(MINYEAR, 1, 1)
DEFAULT_DELTA = timedelta()
DEFAULT_ZONE = UTC
DEFAULT_UUID = uuid.uuid4()

def default_byte():
	"""Initialize the smallest, integer value, i.e. ``Byte``.

	:return: byte
	:rtype: int
	"""
	return int()

def default_character():
	"""Initialize a single, printable character, i.e. ``Character``.

	:return: character
	:rtype: bytes
	"""
	return b' '

def default_rune():
	"""Initialize a single Unicode codepoint, i.e. ``Unicode``.

	:return: codepoint
	:rtype: str
	"""
	return ' '

def default_block():
	"""Initialize a sequence of the smallest integers, i.e. ``Block``.

	:return: fresh instance of an empty block
	:rtype: bytearray
	"""
	return bytearray()  # New object every time.

def default_string():
	"""Initialize a sequence of printable characters, i.e. ``String``.

	:return: an empty sequence of characters
	:rtype: bytes
	"""
	return DEFAULT_STRING

def default_unicode():
	"""Initialize a sequence of Unicode codepoints, i.e. ``Unicode``.

	:return: empty sequence of codepoints
	:rtype: str
	"""
	return DEFAULT_UNICODE

def default_clock():
	"""Initialize a local time variable, i.e. ``ClockTime``.

	:return: the beginning of time
	:rtype: datetime
	"""
	return DEFAULT_CLOCK

def default_span():
	"""Initialize a local time difference variable, i.e. ``TimeSpan``.

	:return: no time
	:rtype: timedelta
	"""
	return DEFAULT_SPAN

def default_world():
	"""Initialize a date-and-time variable, i.e. ``WorldTime``.

	:return: the beginning of time
	:rtype: datetime
	"""
	return DEFAULT_WORLD

def default_delta():
	"""Initialize a date-and-time delta variable, i.e. ``TimeDelta``.

	:return: no time
	:rtype: timedelta
	"""
	return DEFAULT_DELTA

def default_uuid():
	"""Initialize a UUID variable.

	:return: a global, constant UUID value
	:rtype: uuid.UUID
	"""
	return DEFAULT_UUID

def default_array(value, size):
	"""Initialize a vector variable.

	:return: a fresh, empty vector
	:rtype: list
	"""
	return [value] * size   # New object.

def default_vector():
	"""Initialize a vector variable.

	:return: a fresh, empty vector
	:rtype: list
	"""
	return list()   # New object.

def default_set():
	"""Initialize a set variable.

	:return: a fresh, empty set
	:rtype: set
	"""
	return set()

def default_map():
	"""Initialize a map variable.

	:return: a fresh, empty map
	:rtype: dict
	"""
	return dict()

def default_deque():
	"""Initialize a deque variable.

	:return: a fresh, empty double-ended queue
	:rtype: collections.deque
	"""
	return deque()

def default_none():
	return None

# Allow the use of basic python types
# in type expressions.
CONVERT_PYTHON = {
	bool: Boolean(),
	int: Integer8(),
	float: Float8(),
	str: Unicode(),
	bytes: String(),
	bytearray: Block(),
	datetime: WorldTime(),
	timedelta: TimeDelta(),
	uuid.UUID: UUID(),
}

# NEEDS TO BE DROPPED
# Legacy requirement in the arg processing for
# the Utility class.
def fix_expression(a, bread=None):
	"""Promote parameter a from class to instance, as required."""
	if bread is None:
		bread = {}

	if is_portable(a):
		if not is_container(a):
			return a	# No change.
		# Fall thru for structured processing.
	elif is_portable_class(a):
		if not is_container_class(a):
			return a()  # Promotion of simple type.
		raise TypeTrack(a.__name__, 'container class used in type information, instance required')
	elif is_message_class(a):
		return UserDefined(a)
	else:
		# Is it one of the mapped Python classes.
		try:
			e = CONVERT_PYTHON[a]
			return e
		except KeyError:
			pass
		except TypeError:   # Unhashable - list.
			pass
		# Is it an instance of a mapped Python class.
		try:
			e = CONVERT_PYTHON[a.__class__]
			return e
		except KeyError:
			pass
		except AttributeError:   # No class.
			pass
		raise TypeTrack(None, 'not one of the portable types')

	# We have an instance of a structuring.
	try:
		name = a.__class__.__name__
		if isinstance(a, ArrayOf):
			a.element = fix_expression(a.element, bread)
		elif isinstance(a, VectorOf):
			a.element = fix_expression(a.element, bread)
		elif isinstance(a, SetOf):
			a.element = fix_expression(a.element, bread)
		elif isinstance(a, MapOf):
			a.key = fix_expression(a.key, bread)
			a.value = fix_expression(a.value, bread)
		elif isinstance(a, DequeOf):
			a.element = fix_expression(a.element, bread)
		elif isinstance(a, UserDefined):
			if not is_message_class(a.element):
				raise TypeTrack(None, '"%s" is not a user-defined message' % (name,))
		elif isinstance(a, PointerTo):
			try:
				e = bread[id(a)]
			except KeyError:
				e = fix_expression(a.element, bread)
				bread[id(a)] = e
			a.element = e
		else:
			raise TypeTrack(None, 'unexpected container type')
	except TypeTrack as e:
		e.path.append(name)
		raise e
	return a

def override(name, explicit):
	"""Is there explicit information for the named item.

	:param name: name of the message
	:type name: str
	:param explicit: a supplied schema
	:type explicit: a map of <name, portable declaration> pairs
	:return: the explicit information or None
	"""
	if not explicit:
		return None

	try:
		t = explicit[name]
		return t
	except KeyError:
		return None

def infer_type(a):
	"""Map an instance of a Python type to the proper memory description, or None."""
	try:
		t = CONVERT_PYTHON[a.__class__]
		return t
	except AttributeError:
		return None	 # No class.
	except KeyError:
		pass
	if is_message(a):
		return UserDefined(a.__class__)
	return None

def compile_schema(message, explicit_schema):
	"""Produce the best-possible type information for the specified message.

	Use the class and the application-supplied declarations. The
	declarations override any default info that might otherwise be
	acquired from the message.
	"""
	name = message.__name__
	hints = typing.get_type_hints(message)
	class_hints, _ = install_hints(hints)
	if hasattr(message, '__init__'):
		hints = typing.get_type_hints(message.__init__)
		init_hints, init_return = install_hints(hints)
		if init_return:
			raise MessageRegistrationError(name, 'returns value from ctor')

	hint_schema = class_hints or init_hints
	explicit_schema = {k: install_portable(v) for k, v in explicit_schema.items()}

	try:
		m = message()
	except TypeError:
		raise MessageRegistrationError('%s' % (name,), 'not default constructable')
	d = getattr(m, '__dict__', None)
	r = {}
	if d:
		for k, a in d.items():
			t = hint_schema.get(k, None)
			if t is None:
				t = override(k, explicit_schema)
				if t is None:
					t = infer_type(a)

			if not t:
				name_k = f"{name}.{k}"
				reason = 'not enough type information provided/discoverable'
				raise MessageRegistrationError(name_k, reason)
			r[k] = t
	return r

#
def bind_message(message,
		message_trail: bool=True, execution_trace: bool=True,
		copy_before_sending: bool=True, not_portable: bool=False,
		**object_schema):
	"""
	Set the type information and runtime controls for the given message type.

	Values assigned in this function affect the behaviour for all instances of
	the given type.

	:param message: class to be registered as a message
	:type message: :ref:`message<kj-message>`
	:param message_trail: enable log when message is sent
	:param execution_trace: enable log when message is received
	:param copy_before_sending: enable copying of message before each send
	:param not_portable: disable serialization/transfer, e.g. of a file handle
	:param object_schema: explicit type declarations by name
	:rtype: None
	"""
	rt = Runtime(message.__name__, message.__module__,
		message_trail=message_trail,
		execution_trace=execution_trace,
		copy_before_sending=copy_before_sending,
		not_portable=not_portable)

	setattr(message, '__art__', rt)
	if not not_portable:
		rt.schema = compile_schema(message, object_schema)

	install_portable(UserDefined(message))

# This should not need be needed (Incognito) as they are never
# on-the-wire. But registration needed for dispatching within
# encode/decode process.
bind_message(Unknown)
bind_message(Incognito,
	type_name=Unicode,
	decoded_word=Word,
	saved_pointers=MapOf(Unicode, Word),
	address_book=MapOf(Unicode,Address))

#
#
def is_message(m):
	"""Is *m* an instance of a registered class; return a bool."""
	try:
		c = m.__class__
	except AttributeError:
		return False
	b = hasattr(c, '__art__')
	return b

def is_message_class(c):
	"""Has *c* been registered with the library; return a bool."""
	try:
		p = c.__class__	 # Parent class.
	except AttributeError:
		return hasattr(c, '__art__')
	a = getattr(c, '__art__', None)
	b = a is not None and a.name == c.__name__
	return b

class NotFound(object): pass

def equal_to(a, b, t=None, bread=None):
	if bread is None:
		bread = {}
	"""Compare the two operands as instances of portable memory."""
	if t is None:
		if is_message(b):
			if not isinstance(a, b.__class__):
				return False
		elif isinstance(b, Portable):
			a = portable_to_signature(a)
			b = portable_to_signature(b)
			return a == b
		else:
			return a == b
		t = UserDefined(b.__class__)

	if isinstance(t, ArrayOf):
		if len(a) != len(b):
			return False
		return all(equal_to(i, j, t.element, bread) for i, j in zip(a, b))
	elif isinstance(t, VectorOf):
		if len(a) != len(b):
			return False
		return all(equal_to(i, j, t.element, bread) for i, j in zip(a, b))
	elif isinstance(t, DequeOf):
		if len(a) != len(b):
			return False
		return all(equal_to(i, j, t.element, bread) for i, j in zip(a, b))
	elif isinstance(t, SetOf):
		# if len(a) != len(b):
		#	return False
		# return all(i in a for i in b)
		return a == b
	elif isinstance(t, MapOf):
		if len(a) != len(b):
			return False
		#return all(k in b and equal_to(a[k], v, t.value, bread) for k, v in b.items())
		for k, v in b.items():
			x = a.get(k, NotFound)
			if x is NotFound:
				return False
			if not equal_to(x, v, t.value, bread):
				return False
		return True

	elif isinstance(t, UserDefined):
		x = t.element.__art__
		for k, v in x.schema.items():
			try:
				lhs = getattr(a, k)
				rhs = getattr(b, k)
			except AttributeError:
				return False
			if not equal_to(lhs, rhs, v, bread):
				return False
		return True
	elif isinstance(t, Type):
		a = portable_to_signature(a)
		b = portable_to_signature(b)
		return a == b
	elif isinstance(t, PointerTo):
		if a is None:
			if b is None:
				return True
			return False

		i = id(a)
		f = bread.get(i, None)
		if f is not None:
			return True
		bread[i] = True

		i = id(b)
		f = bread.get(i, None)
		if f is not None:
			return True
		bread[i] = True

		return equal_to(a, b, t.element, bread)
	elif isinstance(t, Any):
		if equal_to(a, b, bread=bread):
			return True
		return False
	else:
		return a == b

'''	elif isinstance(p, PointerTo):
		try:
			e = bread[id(p)]
		except KeyError:
			e = convert_portable(p.element, then, bread)
			bread[id(p)] = e
		p.element = e
		then(p.element)
'''


def type_equal_to(a, b):
	"""Compare the two operands as instances of portable memory."""
	if isinstance(a, Portable):
		if isinstance(a, Container):
			pass
		else:
			return a.__class__ == b.__class__

	if isinstance(a, (VectorOf, DequeOf, SetOf, PointerTo)):
		return type_equal_to(a.element, b.element)
	elif isinstance(a, ArrayOf):
		return type_equal_to(a.element, b.element) and a.size == b.size
	elif isinstance(a, MapOf):
		return type_equal_to(a.key, b.key) and type_equal_to(a.value, b.value)
	elif isinstance(a, (Enumeration, UserDefined)):
		return a.element == b.element
	elif isinstance(a, Any):
		return isinstance(b, Any())
	else:
		return a == b



# MIT License
# Copyright (c) 2025 Scott Woods
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Asynchronous distributed computing (OSI layer 7, application)

Repo: git@github.com:mr-ansar/kipjak.git
Branch: main
Commit: f303741b32479857ec90678c3826c930c6eba66b
Version: 
"""

from .general_purpose import *
from .head_lock import *
from .ip_networking import *
from .command_line import *
from .command_startup import *
from .virtual_memory import *
from .convert_memory import *
from .convert_signature import *
from .convert_type import *
from .message_memory import *
from .make_message import *
from .virtual_codec import *
from .json_codec import *
from .object_logs import *
from .virtual_runtime import *
from .object_runtime import *
from .virtual_point import *
from .point_runtime import *
from .routine_point import *
from .point_machine import *
from .file_object import *
from .folder_object import *
from .object_startup import *
from .object_spool import *
from .process_object import *
from .edit_role import *
from .bind_type import *
from .object_directory import *
from .process_directory import *
from .listen_connect import *
from .get_response import *
from .retry_intervals import *
from .disk_storage import *
from .http import *



import socket

__all__ = [
	'get_local_ip',
]

LOCAL_NETWORKS = ('127.', '0:0:0:0:0:0:0:1')
IGNORED_NETWORKS = LOCAL_NETWORKS + ('0.', '0:0:0:0:0:0:0:0', '169.254.', 'fe80:')
TESTNET = '192.0.2.123'

def _get_local_ip(family, remote):
    try:
        s = socket.socket(family, socket.SOCK_DGRAM)
        try:
            s.connect((remote, 9))
            return s.getsockname()[0]
        finally:
            s.close()
    except socket.error:
        return None

def get_local_ip():
	local = _get_local_ip(socket.AF_INET, TESTNET)
	if not local:
		return None

	if local.startswith(IGNORED_NETWORKS):
		return None
	return local



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""kipjak, command line utility.

CRUD for creation/management of a set of process definitions.
Process orchestration for running images of defintiions.
Access to records of execution.
"""
__docformat__ = 'restructuredtext'

import os
import stat
import signal
import re
import datetime
import uuid
import kipjak as kj
import kipjak.rolling_log as rl
import kipjak.process_directory as pd
import kipjak.object_directory as od

from enum import Enum
import calendar

#
GROUP_ROLE = 'group'
GROUP_EXECUTABLE = 'group-jak'

#
def cli(self, *word):
	# At [0] is the tuple created by command_sub_arguments() based on the
	# presence of a valid sub-command name, or its None.
	if not word:
		# A non-sub execution of the utility.
		return kj.Faulted('no sub-command')
	sub = word[0]
	word = word[1:]

	sub_command, jump, sub_args, remainder = sub

	# Catch uncurated args that are not going anywhere.
	if jump not in (create, add, update) and (remainder[0] or remainder[1]):
		a = [k for k in remainder[0].keys()]
		b = [k for k in remainder[1].keys()]
		a.extend(b)
		a = ','.join(a)
		return kj.Faulted(f'cannot execute "{sub_command}"', f'unknown arg(s) "{a}"')

	# Transfer to sub function. Perhaps could be
	# a self.create.
	return jump(self, word, remainder, **sub_args)

kj.bind(cli)

def create(self, word, remainder):
	'''Create a new area on disk to hold process definitions. Return Faulted/None.'''
	home_path = word_i(word, 0) or kj.CL.home_path or kj.DEFAULT_HOME

	cannot_create = f'cannot create "{home_path}"'
	if os.path.isfile(home_path):
		return kj.Faulted(cannot_create, 'existing file')

	elif os.path.isdir(home_path):
		return kj.Faulted(cannot_create, 'existing folder')

	self.console('create', home_path=home_path)

	try:
		f = kj.Folder(home_path)
		f.folder('script')
		f.folder('resource')
		f.folder('role')
		f.folder('departures')
		f.folder('arrivals')

	except OSError as e:
		return kj.Faulted(cannot_create, str(e))

	# Folder created by execution of group create-role.
	self.create(kj.ProcessObject, GROUP_EXECUTABLE,
		origin=kj.ProcessOrigin.RUN,
		home_path=home_path,
		role_name=GROUP_ROLE, top_role=True,
		create_role=True,
		remainder_args=remainder)

	m = self.input()
	if not isinstance(m, kj.Returned):
		return kj.Faulted(cannot_create, f'unexpected process response {m}')

	if isinstance(m.message, kj.Faulted):
		return m.message

	# Consume the expected response.
	if not isinstance(m.message, kj.CommandResponse):
		t = kj.message_to_tag(m.message)
		return kj.Faulted(cannot_create, f'unexpected command response {t}')

	return None

kj.bind(create)

#
#
def add(self, word, remainder, role_count: int=None, role_start: int=0):
	'''Add process definition(s) to an existing store. Return Faulted/None.'''
	executable = word_i(word, 0)
	role_name = word_i(word, 1) or kj.CL.role_name
	home_path = kj.CL.home_path or kj.DEFAULT_HOME

	if executable is None:
		return kj.Faulted('cannot add role', 'no module specified')
	bp = kj.breakpath(executable)

	role_name = role_name or bp[1]
	if role_name is None:
		return kj.Faulted('cannot add role', 'no role specified')

	cannot_add = f'cannot add "{role_name}"'

	if role_name.startswith(GROUP_ROLE):
		return kj.Faulted(cannot_add, 'reserved name')

	home = kj.open_home(home_path)
	if home is None:
		return kj.Faulted(cannot_add, f'home path "{home_path}" does not exist or contains unexpected/incomplete materials')

	if role_count is None:
		role_call = [role_name]
	elif 1 <= role_count <= 1000:
		role_call = [f'{role_name}-{i}' for i in range(role_start, role_start + role_count)]
	else:
		return kj.Faulted(cannot_add, 'expecting role_count in the range 1...1000')

	c = set(home.keys()) & set(role_call)
	if c:
		s = ','.join(c)
		return kj.Faulted(cannot_add, f'collision of roles "{s}"')

	self.console('add', executable=executable, role_name=role_name, home_path=home_path)

	for r in role_call:
		a = self.create(kj.ProcessObject, executable,
			origin=kj.ProcessOrigin.RUN,
			home_path=home_path,
			role_name=r, top_role=True,
			create_role=True,
			remainder_args=remainder)
		m = self.input()
		if not isinstance(m, kj.Returned):
			return kj.Faulted(cannot_add, f'unexpected process response {m}')

		if isinstance(m.message, kj.Faulted):
			return m.message

		if not isinstance(m.message, kj.CommandResponse):
			t = kj.message_to_tag(m.message)
			return kj.Faulted(cannot_add, f'unexpected command response {t}')

	return None

kj.bind(add)

#
#
def list_(self, search, remainder, long_listing: bool=False, group_role: bool=False, sub_roles: bool=False):
	'''List the process definition(s) in an existing store. Return Faulted/None.'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_list = f'cannot list "{home_path}"'

	home = home_listing(self, home_path, search, grouping=group_role, sub_roles=sub_roles)
	if home is None:
		return kj.Faulted(cannot_list, f'does not exist or contains unexpected/incomplete materials')

	s = ','.join(search)
	self.console('list', search=s, home_path=home_path)

	keys = sorted(home.keys())
	for k in keys:
		v = home[k]
		if long_listing:
			home_role = os.path.join(home_path, 'role', k)
			m, _ = kj.storage_manifest(home_role)
			print(f'{k:24} {v.executable_file()} {m.manifests}/{m.listings}/{m.bytes}')
			continue
		print(k)

	return None

kj.bind(list_)

#
#
def update(self, search, remainder):
	'''Update details of existing process definition(s). Return Faulted/None.'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)
	group_role = True
	sub_roles = True

	cannot_update = f'cannot update "{home_path}"'

	home = home_listing(self, home_path, search, grouping=group_role, sub_roles=sub_roles)
	if home is None:
		return kj.Faulted(cannot_update, f'does not exist or contains unexpected/incomplete materials')

	r = ','.join(home.keys())
	self.console('update', roles=r, home_path=home_path)

	for k, v in home.items():
		a = self.create(kj.ProcessObject, v,
			origin=kj.ProcessOrigin.RUN,
			home_path=home_path,
			role_name=k, top_role=True,
			update_role=True,
			remainder_args=remainder)
		m = self.input()
		if not isinstance(m, kj.Returned):
			return kj.Faulted(cannot_update, f'unexpected process response {m}')

		if isinstance(m.message, kj.Faulted):
			return m.message

		if not isinstance(m.message, kj.CommandResponse):
			t = kj.message_to_tag(m.message)
			return kj.Faulted(cannot_update, f'not a proper command response {t}')

	return None

kj.bind(update)

#
#
def delete(self, search, remainder, all_roles: bool=False):
	'''Delete existing process definition(s). Return Faulted/None.'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_delete = f'cannot delete from "{home_path}"'
	if search:
		home = home_listing(self, home_path, search)
	elif all_roles:
		home = kj.open_home(home_path)
		if home is None:
			return kj.Faulted(cannot_delete, f'does not exist or unexpected/incomplete materials')
	else:
		return kj.Faulted(cannot_delete, f'no roles specified, use --all_roles?')

	self.console('delete', search=search, home_path=home_path)

	try:
		running = home_running(self, home)
		if running:
			r = ','.join(running.keys())
			return kj.Faulted(cannot_delete, f'roles "{r}" are currently running')

	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	try:
		role_path = os.path.join(home_path, 'role')
		for p in os.listdir(role_path):
			if p.startswith('group'):
				continue
			s = p.split('.')
			if s[0] not in home:
				continue
			home_role = os.path.join(role_path, p)
			kj.remove_folder(home_role)
	except FileNotFoundError as e:
		return kj.Faulted(cannot_delete, str(e))

	return None

kj.bind(delete)

#
#
def destroy(self, word, remainder):
	'''Remove all trace of an existing store. Return Faulted/None.'''
	home_path = word_i(word, 0) or kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_destroy = f'cannot destroy "{home_path}"'

	# Most basic checks.
	if os.path.isfile(home_path):
		return kj.Faulted(cannot_destroy, f'existing file')

	elif not os.path.isdir(home_path):
		return kj.Faulted(cannot_destroy, f'not a folder')

	# Need valid contents to be able to check
	# process status.
	home = kj.open_home(home_path)
	if home is None:
		return kj.Faulted(cannot_destroy, f'does not exist or unexpected/incomplete materials')

	self.console('destroy', home_path=home_path)

	try:
		running = home_running(self, home)
		if not running:
			kj.remove_folder(home_path)
	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	if running:
		r = ','.join(running.keys())
		return kj.Faulted(cannot_destroy, f'running roles "{r}"')

	return None

kj.bind(destroy)

#
#
def run(self, search, remainder, main_role: str=None):
	'''Execute a subset/all of the process definition(s), to completion. Return Faulted/None.'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_run = f'cannot run "{home_path}"'
	if search:
		home = home_listing(self, home_path, search)
	else:
		home = kj.open_home(home_path)
		if not home:
			return kj.Faulted(cannot_run, f'does not exist or unexpected/incomplete materials')

	s = ','.join(search)
	self.console('run', search=s, home_path=home_path)

	try:
		kv = {}
		if main_role is not None:
			kv['main_role'] = main_role

		running = home_running(self, home)
		if not running:
			a = self.create(kj.ProcessObject, GROUP_EXECUTABLE, *search,
				home_path=home_path,
				role_name=GROUP_ROLE, top_role=True,
				origin=kj.ProcessOrigin.RUN,
				**kv)
			self.assign(a, 0)
			m = self.input()
			if isinstance(m, kj.Returned):
				self.debrief()
				return m.message
			elif isinstance(m, kj.Faulted):
				return m
			elif isinstance(m, kj.Stop):
				return kj.Aborted()

	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	if running:
		r = ','.join(running.keys())
		return kj.Faulted(cannot_run, f'roles "{r}" are already running')

	return None

kj.bind(run)

#
#
def start(self, search, remainder, main_role: str=None):
	'''Start a subset/all of the process definition(s) as background daemons. Return Faulted/None (immediately).'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_start = f'cannot start "{home_path}"'
	if search:
		home = home_listing(self, home_path, search)
	else:
		home = kj.open_home(home_path)
		if home is None:
			return kj.Faulted(cannot_start, f'does not exist or unexpected/incomplete materials')

	if not home:
		return kj.Faulted(cannot_start, f'empty')

	s = ','.join(search)
	self.console('start', search=s, home_path=home_path)

	try:
		kv = {}
		if main_role is not None:
			kv['main_role'] = main_role

		running = home_running(self, home)
		if not running:
			a = self.create(kj.ProcessObject, GROUP_EXECUTABLE, *search,
				home_path=home_path,
				role_name=GROUP_ROLE, top_role=True,
				origin=kj.ProcessOrigin.START,
				**kv)
			self.assign(a, 0)
			m = self.input()
			if isinstance(m, kj.Returned):
				self.debrief()
				if not isinstance(m.message, kj.CommandResponse):
					return kj.Faulted(cannot_start, f'unexpected response from group')
				return None
			elif isinstance(m, kj.Faulted):
				return m
			elif isinstance(m, kj.Stop):
				return kj.Aborted()

	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	if running:
		r = ','.join(running.keys())
		return kj.Faulted(cannot_start, f'roles "{r}" are already running')

	return None

kj.bind(start)

#
#
def stop(self, search, remainder):
	'''Stop the previously started set of background daemons. Return Faulted/None (immediately).'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_stop = f'cannot stop "{home_path}"'
	home = kj.open_home(home_path)
	if home is None:
		return kj.Faulted(cannot_stop, f'does not exist or unexpected/incomplete materials')

	if not home:
		return kj.Faulted(cannot_stop, f'empty')

	self.console('stop', home_path=home_path)

	try:
		running = home_running(self, home)
		if not running:
			return kj.Faulted(cannot_stop, f'nothing running')
		parent_pid = set()

		for k, v in running.items():
			parent_pid.add(v.parent_pid)

		for p in parent_pid:
			os.kill(p, signal.SIGINT)

	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	return None

kj.bind(stop)

#
#
def status(self, search, remainder, long_listing: bool=False, group_role: bool=False, sub_roles: bool=False):
	'''Query running/not-running status of background daemons. Return Faulted/None.'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_status = f'cannot query status "{home_path}"'

	# Get list of roles at home_path, trimmed down
	# according to the list of search patterns.
	home = home_listing(self, home_path, search, grouping=group_role, sub_roles=sub_roles)
	if home is None:
		return kj.Faulted(cannot_status, f'does not exist or unexpected/incomplete materials')

	if not home:
		return kj.Faulted(cannot_status, f'empty')

	s = ','.join(search)
	self.console('status', search=s, home_path=home_path)

	try:
		# Determine the running/idle status of the
		# selected roles.
		running = home_running(self, home)
		if not running:
			return kj.Faulted(cannot_status, f'nothing running')

	finally:
		# Cleanup the locking from inside home_running().
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	now = datetime.datetime.now(kj.UTC)
	orderly = sorted(running.keys())

	def long_status():
		for k in orderly:
			v = home.get(k, None)
			r = running.get(k, None)

			start_stop = v.start_stop()
			s = start_stop[-1]
			if s.start is not None:
				d = now - s.start
				s = short_delta(d)
			else:
				s = '(never started)'

			def dq(pid):
				if pid is None:
					return '0'
				return f'{pid}'
			kj.output_line('%-24s <%s> %s' % (k, dq(r.pid), s))

	def short_status():
		for k in orderly:
			kj.output_line(k)

	if long_listing:
		long_status()
	else:
		short_status()

	return None

kj.bind(status)

#
#
def history(self, word, remainder, long_listing: bool=False):
	'''List the the start/stop record for the specified role. Return Faulted/None.'''
	role_name = word_i(word, 0) or kj.CL.role_name
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	group_role = True
	sub_roles = True

	if role_name is None:
		return kj.Faulted(f'cannot pull history', f'no role specified')
	cannot_history = f'cannot pull history for "{role_name}"'

	home = kj.open_home(home_path, grouping=group_role, sub_roles=sub_roles)
	if home is None:
		return kj.Faulted(cannot_history, f'home at "{home_path}" does not exist or contains unexpected/incomplete materials')

	role = home.get(role_name, None)
	if role is None:
		return kj.Faulted(cannot_history, f'does not exist')

	self.console('history', role_name=role_name, home_path=home_path)

	def long_history():
		now = datetime.datetime.now(kj.UTC)
		for s in role.start_stop():
			start = kj.world_to_text(s.start)
			if s.stop is None:
				kj.output_line('%s ... ?' % (start,))
				continue
			stop = kj.world_to_text(s.stop)
			d = s.stop - s.start
			span = '%s' % (kj.short_delta(d),)
			if isinstance(s.returned, kj.Incognito):
				kj.output_line('%s ... %s (%s) %s' % (start, stop, span, s.returned.type_name))
			else:
				kj.output_line('%s ... %s (%s) %s' % (start, stop, span, s.returned.__class__.__name__))

	def short_history():
		for i, s in enumerate(role.start_stop()):
			now = datetime.datetime.now(kj.UTC)
			d = now - s.start
			start = '%s ago' % (kj.short_delta(d),)
			if s.stop is None:
				kj.output_line('[%d] %s ... ?' % (i, start))
				continue
			d = s.stop - s.start
			stop = kj.short_delta(d)
			if isinstance(s.returned, kj.Incognito):
				kj.output_line('[%d] %s ... %s (%s)' % (i, start, stop, s.returned.type_name))
			else:
				kj.output_line('[%d] %s ... %s (%s)' % (i, start, stop, s.returned.__class__.__name__))

	if long_listing:
		long_history()
	else:
		short_history()
	return None

kj.bind(history)

#
#
class NoFault(object):
	def __init__(self, fault: kj.Any=None):
		self.fault = fault

kj.bind(NoFault)

def returned(self, word, remainder, start: int=None, timeout: float=None):
	'''Print a specific termination value for the specified role. Return Faulted/None.'''
	role_name = word_i(word, 0) or kj.CL.role_name
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	group_role = True
	sub_roles = True

	if role_name is None and group_role:
		role_name = GROUP_ROLE

	if role_name is None:
		return kj.Faulted(f'cannot pull return', f'no role specified')
	cannot_returned = f'cannot pull return for "{role_name}"'

	home = kj.open_home(home_path, grouping=group_role, sub_roles=sub_roles)
	if home is None:
		return kj.Faulted(cannot_returned, f'home at "{home_path}" does not exist or contains unexpected/incomplete materials')

	role = home.get(role_name, None)
	if role is None:
		return kj.Faulted(cannot_returned, f'does not exist')

	self.console('returned', role_name=role_name, home_path=home_path)

	start_stop = role.start_stop()
	if len(start_stop) < 1:
		return kj.Faulted(cannot_returned, f'no start/stop records')

	ls = len(start_stop) - 1	# Last stop
	if start is None:
		start = ls
	elif start < 0:
		start = ls + 1 + start

	if start < 0 or start > ls:
		return kj.Faulted(cannot_returned, f'start [{start}] out of bounds')

	# Criteria met - valid row in the table.
	selected = start_stop[start]
	anchor = selected.start

	def no_fault(value):
		if isinstance(value, kj.Faulted):
			return NoFault(value)
		return selected.returned

	# This row has already returned.
	if selected.stop is not None:
		return no_fault(selected.returned)

	# Cannot poll for completion of anything other
	# than the last row.
	if start != ls:
		return kj.Faulted(cannot_returned, f'no record of role[{start}] stopping and never will be')

	if timeout is not None:
		self.start(kj.T1, timeout)

	self.start(kj.T2, 1.0)
	while True:
		m, i = self.select(kj.Stop, kj.T1, kj.T2)
		if isinstance(m, kj.Stop):
			break
		elif isinstance(m, kj.T1):
			return kj.TimedOut(m)
		elif isinstance(m, kj.T2):
			r = role.start_stop.resume()
			if len(r) < start:
				return kj.Faulted(cannot_returned, f'lost original start position')
			if r[start].start != anchor:
				return kj.Faulted(cannot_returned, f'lost original start position, datetime anchor')
			if r[start].stop is not None:
				return no_fault(r[start].returned)
			self.start(kj.T2, 1.0)

	return None

kj.bind(returned)

#
#
# Extraction of logs for a role.
#
class TimeFrame(Enum):
	"""
	Enumeration of a general-usage time unit.

	* MONTH - 4 weeks
	* WEEK - 7 days
	* DAY - 24 hours
	* HOUR - 60 minutes
	* MINUTE - 60 seconds
	* HALF - 30 minutes
	* QUARTER - 15 minutes
	* TEN - 10 minutes
	* FIVE - 5 minutes
	"""
	MONTH=0
	WEEK=1
	DAY=2
	HOUR=3
	MINUTE=4
	HALF=5
	QUARTER=6
	TEN=7
	FIVE=8

def log(self, word, remainder, clock: bool=False,
	tail: int=None, from_: str=None, last: TimeFrame=None, start: int=None, back=None,
	to: str=None, span=None, count: int=None, sample: str=None, tags: str=None):
	'''List logging records for the specified process definition. Return Faulted/None.'''
	role_name = word_i(word, 0) or kj.CL.role_name
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	group_role = True
	sub_roles = True

	if role_name is None:
		return kj.Faulted(f'cannot log', f'no role specified')
	cannot_log = f'cannot log "{role_name}"'

	self.console('log', role_name=role_name, home_path=home_path)

	# Need a beginning and an ending.
	f = [
		tail,		# Rewind up to a specified number of lines.
		from_,		# At a specific datetime.
		last,		# Recent timeframe.
		start,		# Entry in the start-stop log.
		back		# Timespan back from now.
	]
	c = len(f) - f.count(None)
	if c == 0:
		tail = os.get_terminal_size().lines - 2
	elif c != 1:
		# one of <tail>, <from>, <last>, <start> or <back> is required
		return kj.Faulted(cannot_log, f'need one of tail, from_, last, start or back')

	# The ending (optional).
	t = [
		to,			# Specific datetime.
		span,		# Timespan from start.
		count		# Number of lines from start.
	]
	c = len(t) - t.count(None)
	if c == 0:
		pass		# Default is query to end-of-log.
	elif c != 1:
		# one of <to>, <span> or <count> is required
		return kj.Faulted(cannot_log, f'optional use of to, span or count')

	# Open the home and role.
	home = kj.open_home(home_path, grouping=group_role, sub_roles=sub_roles)
	if home is None:
		return kj.Faulted(cannot_log, f'home at "{home_path}" does not exist or contains unexpected/incomplete materials')

	role = home.get(role_name, None)
	if role is None:
		return kj.Faulted(cannot_log, f'does not exist')

	# Calculate the start position and then the
	# end position - might be relative to start.
	begin, end = None, None
	if tail is not None:
		if tail < 1:
			return kj.Faulted(cannot_log, f'tail [{tail}] out of range')

	elif from_ is not None:
		begin = world_or_clock(from_, clock)

	elif last is not None:
		begin = from_last(last)
		if begin is None:
			return kj.Faulted(cannot_log, f'last is not a TimeFrame')

	elif start is not None:
		start_stop = role.start_stop()
		if len(start_stop) < 1:
			return kj.Faulted(cannot_log, f'no history available')
		if start < 0:
			y = len(start_stop) + start
		else:
			y = start
		try:
			s = start_stop[y]
		except IndexError:
			return kj.Faulted(cannot_log, f'start [{y} out of range]')
		begin = s.start
		p1 = y + 1
		if p1 < len(start_stop):
			end = start_stop[p1].start
		else:
			end = None
	elif back is not None:
		d = datetime.datetime.now(kj.UTC)
		t = datetime.timedelta(seconds=back)
		begin = d - t

	# Calculate the end.
	if to is not None:
		end = world_or_clock(to, clock)
	elif span is not None:
		#t = datetime.timedelta(seconds=span)
		if begin is None:
			return kj.Faulted(cannot_log, f'span has nowhere to start')
		end = begin + span

	# Now that <begin>(or tail) and <end> have been established, a
	# few more sanity checks.
	if begin is None and tail is None:
		return kj.Faulted(cannot_log, f'<begin> not defined and not inferred')

	if begin is not None and end is not None and end < begin:
		return kj.Faulted(cannot_log, f'<end> comes before <begin>')

	if sample is not None and tags is not None:
		return kj.Faulted(cannot_log, f'sampling is by a specific tag')

	# Boundaries are set. Ready to scan.
	if sample:
		a = self.create(sampler, role, clock, begin, tail, end, count, sample)
	else:
		a = self.create(printer, role, clock, begin, tail, end, count, tags)

	m, i = self.select(kj.Stop, kj.Returned)
	if isinstance(m, kj.Stop):
		kj.halt(a)
		m, i = self.select(kj.Returned)
		return kj.Aborted()

	message = m.message
	if message is None:   # Reached the end.
		pass
	elif isinstance(message, kj.Faulted):	 # kj.Failed to complete stream.
		return message
	else:
		return kj.Faulted(cannot_log, f'unexpected reader response')

	return None

kj.bind(log, span=kj.TimeSpan(), back=kj.TimeSpan())

#
#
def edit(self, word, remainder):
	'''Edit the configuration of the specified process defintion. Return Faulted/None.'''
	role_name = word_i(word, 0) or kj.CL.role_name
	home_path = word_i(word, 1) or kj.CL.home_path or kj.DEFAULT_HOME
	group_role = True
	sub_roles = True

	if role_name is None:
		return kj.Faulted(f'cannot edit "{home_path}"', f'no role specified')

	cannot_edit = f'cannot edit "{role_name}"'

	home = kj.open_home(home_path, grouping=group_role, sub_roles=sub_roles)
	if home is None:
		return kj.Faulted(cannot_edit, f'does not exist or contains unexpected/incomplete materials')

	try:
		running = home_running(self, home)
		if role_name in running:
			return kj.Faulted(cannot_edit, f'role is running')

	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	r = home.get(role_name, None)
	if r is None:
		return kj.Faulted(cannot_edit, f'does not exist')

	self.console('edit', role_name=role_name, home_path=home_path)

	output = kj.HR.edit_role(self, r)
	return output

kj.bind(edit)

#
#
def print_scope(s):
	s = str(s)
	i = s.rfind('.')
	return s[i + 1:]

def print_id(uid, full_identity=False):
	s = str(uid)
	if full_identity:
		return s
	return s[:8]

#
#
def print_network(self, d, ipp, full_identity=False, directory_addresses=False, tab=0,
		list_published=False, list_subscribed=False, list_routed=False, list_connected=False):

	#kj.output_line(f'scope={d.scope}, unique_id={d.unique_id}, subs={len(d.sub_directory)}')
	#for k, v in d.sub_directory.items():
	#	kj.output_line(f'k={k}, v={v}')

	i = print_id(d.unique_id, full_identity=full_identity)
	s = print_scope(d.scope)
	a = str(ipp)

	if directory_addresses:
		kj.output_line(f'[{s}] {d.executable} ({i}) {d.primary_ip} <C>{d.connect_to_directory} <L>{d.accept_directories_at}', tab=tab)
	else:
		kj.output_line(f'[{s}] {d.executable} ({i})', tab=tab)

	if list_published:
		for p in d.published:
			t = print_id(p.published_id, full_identity=full_identity)
			ipp = p.listening_ipp
			if ipp and ipp.host:
				kj.output_line(f'# "{p.name}" ({t}) {p.listening_ipp}', tab=tab+1)
			else:
				kj.output_line(f'# "{p.name}" ({t})', tab=tab+1)

	if list_subscribed:
		for s in d.subscribed:
			t = print_id(s.subscribed_id, full_identity=full_identity)
			kj.output_line(f'? "{s.search}" ({t})', tab=tab+1)

	if list_routed:
		for r in d.routed:
			a = print_id(r.subscribed_id, full_identity=full_identity)
			b = print_id(r.published_id, full_identity=full_identity)
			kj.output_line(f'<>" {r.name}" ({a} -> {b})', tab=tab+1)

	if list_connected:
		for subscribed_id, dc in d.peer.items():
			a = print_id(subscribed_id, full_identity=full_identity)
			kj.output_line(f'? "{dc.search}" ({a})', tab=tab+1)
			for ds in dc.session:
				route = ds.route
				scope = print_scope(route.scope)
				if route is None:
					kj.output_line(f'> "{ds.name}" (no active route)', tab=tab+2)
				elif isinstance(route, od.RouteOverLoop):
					if isinstance(ds.status, kj.Connected):
						requested_ipp = ds.status.request.requested_ipp
						opened_ipp = ds.status.opened_ipp
						kj.output_line(f'> "{ds.name}"[{scope}] ({opened_ipp} -> {requested_ipp})', tab=tab+2)
					elif isinstance(ds.status, kj.Faulted):
						kj.output_line(f'"{ds.name}"[{scope}] {route.ipp} ({ds.status})', tab=tab+2)
					else:
						tag = kj.message_to_tag(ds.status)
						kj.output_line(f'> "{ds.name}"[{scope}] {route.ipp} (unexpected "{tag}")', tab=tab+2)
				elif isinstance(route, od.RouteToAddress):
					pi = print_id(route.published_id, full_identity=full_identity)
					si = print_id(route.subscribed_id, full_identity=full_identity)
					kj.output_line(f'> "{ds.name}"[{scope}] ({si} -> {pi})', tab=tab+2)

	for k, v in d.sub_directory.items():
		self.send(od.GetDirectory(), v)
		m = self.input()
		if isinstance(m, od.DirectoryListing):
			print_network(self, m, k, full_identity=full_identity, directory_addresses=directory_addresses, tab=tab+1,
				list_published=list_published, list_subscribed=list_subscribed, list_routed=list_routed, list_connected=list_connected)
		elif isinstance(m, kj.Faulted):
			return m
		elif isinstance(m, kj.Stop):
			return kj.Aborted()

	return None

def network(self, word, remainder, open_scope: kj.ScopeOfDirectory=None, full_identity: bool=False, directory_addresses: bool=False,
		list_published: bool=False, list_subscribed: bool=False, list_routed: bool=False, list_connected: bool=False):
	'''. Return Faulted/None.'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME

	if home_path is None:
		return kj.Faulted(f'cannot edit "{home_path}"', f'no role specified')
	open_scope = open_scope or kj.ScopeOfDirectory.HOST

	# Check for connected directory.
	self.send(od.OpenDirectory(scope=open_scope), pd.PD.directory)
	self.start(kj.T1, 30.0)
	while True:
		m = self.input()
		if isinstance(m, od.DirectoryOpened):
			break
		elif isinstance(m, kj.Faulted):
			return m
		elif isinstance(m, kj.T1):
			return kj.TimedOut(m)
		elif isinstance(m, kj.Stop):
			return kj.Aborted()

	# Directory is connected to something. Query
	# for the whole tree.
	self.send(od.ListDirectory(), pd.PD.directory)
	self.start(kj.T2, 10.0)

	m = self.input()
	if isinstance(m, od.DirectoryListing):
		output = print_network(self, m, 'root',
			full_identity=full_identity, directory_addresses=directory_addresses,
			list_published=list_published, list_subscribed=list_subscribed,
			list_routed=list_routed, list_connected=list_connected)
	elif isinstance(m, kj.Faulted):
		return m
	elif isinstance(m, kj.T2):
		return kj.TimedOut(m)
	elif isinstance(m, kj.Stop):
		return kj.Aborted()

	return output

kj.bind(network)

#
#
def find_name(self, d, name):
	s = str(d.unique_id)
	if s.startswith(name):
		return d

	for p in d.published:
		t = str(p.published_id)
		if t.startswith(name):
			return p.published_id

	for s in d.subscribed:
		t = str(s.subscribed_id)
		if t.startswith(name):
			return s.subscribed_id

	for k, v in d.sub_directory.items():
		self.send(od.GetDirectory(), v)
		m = self.input()
		if isinstance(m, od.DirectoryListing):
			f = find_name(self, m, name)
			if f:
				return f
		elif isinstance(m, kj.Faulted):
			return m
		elif isinstance(m, kj.Stop):
			return kj.Aborted()

	return None

def ping(self, word, remainder, open_scope: kj.ScopeOfDirectory=None, ping_count: int=4):
	'''. Return Faulted/None.'''
	name = word_i(word, 0)

	if name is None:
		return kj.Faulted(f'cannot ping', f'no identity specified')
	open_scope = open_scope or kj.ScopeOfDirectory.HOST

	# Check for connected directory.
	self.send(od.OpenDirectory(scope=open_scope), pd.PD.directory)
	self.start(kj.T1, 30.0)
	while True:
		m = self.input()
		if isinstance(m, od.DirectoryOpened):
			break
		elif isinstance(m, kj.Faulted):
			return m
		elif isinstance(m, kj.T1):
			return kj.TimedOut(m)
		elif isinstance(m, kj.Stop):
			return kj.Aborted()

	# Directory is connected to something. Query
	# for the whole tree.
	self.send(od.ListDirectory(), pd.PD.directory)
	self.start(kj.T2, 10.0)
	while True:
		m = self.input()
		if isinstance(m, od.DirectoryListing):
			break
		elif isinstance(m, kj.Faulted):
			return m
		elif isinstance(m, kj.T2):
			return kj.TimedOut(m)
		elif isinstance(m, kj.Stop):
			return kj.Aborted()

	d = find_name(self, m, name)
	if d is None:
		return kj.Faulted(f'cannot ping "{name}"', f'not present in directory')
	elif isinstance(d, uuid.UUID):
		return kj.Faulted(f'cannot ping "{name}"', f'not a directory node')
	elif isinstance(d, od.DirectoryListing):
		pass
	else:
		return kj.Faulted(f'cannot ping "{name}"', f'unexpected object')

	for i in range(ping_count):
		if i:
			self.start(kj.T2, 1.0)

			m = self.input()
			if isinstance(m, kj.T2):
				pass
			elif isinstance(m, kj.Stop):
				return kj.Aborted()

		kj.output_line(f'[{i}] ... ', newline=False)

		b = kj.clock_now()
		self.send(kj.Ping(), d.directory_address)
		self.start(kj.T1, 3.0)

		m = self.input()
		if isinstance(m, kj.Ping):
			e = kj.clock_now()
			t = e - b
			t = kj.span_to_text(t)
			kj.output_line(f'{t}')
		elif isinstance(m, kj.T1):
			kj.output_line(f'(timed out)')
		elif isinstance(m, kj.Stop):
			return kj.Aborted()

	output = None
	return output

kj.bind(ping)

#
#
def list_folder(path, recursive_listing):
	for s in os.listdir(path):
		p = os.path.join(path, s)
		if os.path.isdir(p):
			yield p
			if recursive_listing:
				yield from list_folder(p, True)
		elif os.path.isfile(p):
			yield p

def get_printer(target_path, full_path, long_listing):
	if full_path:
		if long_listing:
			def printer(path):
				st = os.stat(path)
				fm = stat.filemode(st.st_mode)
				print(f'{fm} {st.st_uid}/{st.st_gid} {path}')
		else:
			def printer(path):
				print(path)
	else:
		h = len(target_path)
		if long_listing:
			def printer(path):
				st = os.stat(path)
				fm = stat.filemode(st.st_mode)
				hd = path[h + 1:]
				print(f'{fm} {st.st_uid}/{st.st_gid} {hd}')
		else:
			def printer(path):
				hd = path[h + 1:]
				print(hd)
	return printer

def resource(self, word, remainder,
		full_path: bool=False, recursive_listing: bool=False, long_listing: bool=False,
		make_changes: bool=False, clear_all: bool=False):
	'''.'''
	if not word:
		return kj.Faulted(f'cannot resource group', f'no executable specified')

	executable = word[0]
	word = word[1:]
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_resource = f'cannot resource "{executable}"'

	home = kj.open_home(home_path, grouping=True, sub_roles=True)
	if home is None:
		return kj.Faulted(cannot_resource, f'does not exist or contains unexpected/incomplete materials')

	def matching(executable_file):
		s = os.path.split(executable_file)
		return s[1] == executable

	role_matching = {k: v for k, v in home.items() if matching(v.executable_file())}

	try:
		running = home_running(self, role_matching)
		n = len(running)
		if n > 1:
			r = ','.join(running.keys())
			return kj.Faulted(cannot_resource, f'roles "{r}" are running')
		elif n == 1:
			r = next(iter(running.keys()))
			return kj.Faulted(cannot_resource, f'role "{r}" is running')

	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	try:
		resource_path = kj.CL.resource_path
		target_path = os.path.join(home_path, 'resource', executable)
		if not os.path.isdir(target_path):
			return kj.Faulted(cannot_resource, f'folder "{target_path}" does not exist')

		if not word:
			if resource_path:
				source_storage, _ = kj.storage_manifest(resource_path)
				target_storage, _ = kj.storage_manifest(target_path)
			elif clear_all:
				kj.remove_contents(target_path)
				return None
			else:
				printer = get_printer(target_path, full_path, long_listing)
				for r in list_folder(target_path, recursive_listing):
					printer(r)
				return None
		else:
			if resource_path or clear_all or full_path or recursive_listing:
				return kj.Faulted(cannot_resource, 'inappropriate argument(s)')
			source_storage, _ = kj.storage_selection(word, path=os.getcwd())
			target_storage, _ = kj.storage_manifest(target_path)

		storage_delta = [d for d in kj.storage_delta(source_storage, target_storage)]

		if not storage_delta:			# Nothing to see or do.
			return None

		if not make_changes:			# Without explicit command, show what would happen.
			for d in storage_delta:
				print(d)
			return None

		a = self.create(kj.FolderTransfer, storage_delta, target_storage.path)

		m, _ = self.select(kj.Returned, kj.Stop)
		if isinstance(m, kj.Stop):
			self.send(m, a)
			m, _ = self.select(kj.Returned)
			return kj.Aborted()

	except (OSError, ValueError) as e:
		return kj.Faulted(cannot_resource, str(e))

	message = m.message
	if isinstance(message, kj.Faulted):
		return message
	return None

kj.bind(resource)

#
#
def model(self, word, remainder,
		full_path: bool=False, recursive_listing: bool=False, long_listing: bool=False,
		make_changes: bool=False, clear_all: bool=False,
		get_latest: str=None):
	'''.'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_model = f'cannot model "{home_path}"'

	if not word:
		return kj.Faulted(cannot_model, f'no role specified')
	role = word[0]
	word = word[1:]

	cannot_model = f'cannot model "{role}"'

	home = kj.open_home(home_path, grouping=True, sub_roles=True)
	if home is None:
		return kj.Faulted(cannot_model, f'does not exist or contains unexpected/incomplete materials')

	# Can match only 1. Keep consistent code
	# layout for status checks.
	def matching(k):
		return k == role

	role_matching = {k: v for k, v in home.items() if matching(k)}

	if not role_matching:
		return kj.Faulted(cannot_model, f'unknown role')

	model_path = kj.CL.model_path
	try:
		running = home_running(self, role_matching)
		n = len(running)
		if word or model_path or clear_all or make_changes:
			if n > 1:
				r = ','.join(running.keys())
				return kj.Faulted(cannot_model, f'roles "{r}" are running')
			elif n == 1:
				r = next(iter(running.keys()))
				return kj.Faulted(cannot_model, f'role "{r}" is running')

	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	if model_path or clear_all or full_path or recursive_listing:
		return kj.Faulted(cannot_model, 'inappropriate argument(s)')

	try:
		target_path = os.path.join(home_path, 'role', role, 'model')
		if not os.path.isdir(target_path):
			return kj.Faulted(cannot_model, f'role model folder is not usable')

		if not word:
			if model_path:
				# Folder-based asset management.
				model_path = os.path.abspath(model_path)
				if not os.path.isdir(model_path):
					return kj.Faulted(cannot_model, f'path "{model_path}" is not a usable folder')

				source_storage, _ = kj.storage_manifest(model_path)
				target_storage, _ = kj.storage_manifest(target_path)

			elif get_latest:
				# Retrieve assets from the role.
				get_latest = os.path.abspath(get_latest)
				if not os.path.exists(get_latest):
					kj.Folder(get_latest)
				elif not os.path.isdir(get_latest):
					return kj.Faulted(cannot_model, f'path "{get_latest}" is not a usable folder')

				source_storage, _ = kj.storage_manifest(target_path)
				target_storage, _ = kj.storage_manifest(get_latest)

			elif clear_all:
				kj.remove_contents(target_path)
				return None
			else:
				printer = get_printer(target_path, full_path, long_listing)
				for r in list_folder(target_path, recursive_listing):
					printer(r)
				return None
		else:
			if get_latest:
				get_latest = os.path.abspath(get_latest)
				if not os.path.exists(get_latest):
					kj.Folder(get_latest)
				elif not os.path.isdir(get_latest):
					return kj.Faulted(cannot_model, f'path "{get_latest}" is not a usable folder')

				source_storage, _ = kj.storage_manifest(target_path)
				target_storage, _ = kj.storage_selection(word, path=get_latest)
			else:
				source_storage, _ = kj.storage_selection(word)
				target_storage, _ = kj.storage_manifest(target_path)

		storage_delta = [d for d in kj.storage_delta(source_storage, target_storage)]

		if not storage_delta:			# Nothing to see or do.
			return None

		if not make_changes:			# Without explicit command, show what would happen.
			for d in storage_delta:
				print(d)
			return None

		a = self.create(kj.FolderTransfer, storage_delta, target_storage.path)

		m, _ = self.select(kj.Returned, kj.Stop)
		if isinstance(m, kj.Stop):
			self.send(m, a)
			m, _ = self.select(kj.Returned)
			return kj.Aborted()

	except (OSError, ValueError) as e:
		return kj.Faulted(cannot_model, str(e))

	message = m.message
	if isinstance(message, kj.Faulted):
		return message
	return None

kj.bind(model)

#
#
def script(self, word, remainder,
		full_path: bool=False, recursive_listing: bool=False, long_listing: bool=False,
		list_scripts: bool=False, list_executables: bool=False, list_paths: bool=False,
		make_changes: bool=False, clear_all: bool=False):
	'''.'''
	home_path = kj.CL.home_path or kj.DEFAULT_HOME
	home_path = os.path.abspath(home_path)

	cannot_script = f'cannot script "{home_path}"'

	home = kj.open_home(home_path, grouping=True, sub_roles=True)
	if home is None:
		return kj.Faulted(cannot_script, f'does not exist or contains unexpected/incomplete materials')

	try:
		running = home_running(self, home)
		n = len(running)
		if n and (clear_all or make_changes):
			r = ','.join(running.keys())
			return kj.Faulted(cannot_script, f'roles "{r}" are running')

	finally:
		self.abort()
		while self.working():
			m, i = self.select(kj.Returned)
			self.debrief()

	# Get all executable for every role, then refine that to a map of
	# unique paths, i.e. where each python module comes from.
	role_executable = {k: v.executable_file() for k, v in home.items()}
	source_path = {os.path.split(v)[0] for k, v in role_executable.items() if v.endswith('.py')}

	if not source_path:
		return kj.Faulted(cannot_script, f'no scripts in use')

	# Combine the file and folder names from each unique path into
	# a single collection.
	selection = []
	collision = []
	for p in source_path:
		for s in os.listdir(p):
			# Need special handling of "library" module. Allow
			# the first one through. If present at this level its
			# intended that they are all empty.
			if s == '__init__.py':
				if s in selection:
					continue
			elif s.startswith('__') and s.endswith('__'):
				# Skip any other special python materials, e.g. cache.
				continue
			elif s in selection:
				collision.append(s)
				continue
			t = os.path.join(p, s)
			selection.append(t)

	if collision:
		c = ','.join(collision)
		return kj.Faulted(cannot_script, f'duplicated names "{c}"')

	try:
		target_path = os.path.join(home_path, 'script')

		listing = list_scripts or list_executables or list_paths
		if not listing and not clear_all:
			if full_path or long_listing or recursive_listing:
				return kj.Faulted(cannot_script, 'inappropriate argument(s)')

			source_storage, _ = kj.storage_selection(selection, path=os.getcwd())
			target_storage, _ = kj.storage_manifest(target_path)
		elif list_scripts:
			printer = get_printer(target_path, full_path, long_listing)
			for r in list_folder(target_path, recursive_listing):
				printer(r)
			return None
		elif list_executables:
			for k, v in role_executable.items():
				print(f'{k:24} {v}')
			return None
		elif list_paths:
			for s in source_path:
				print(f'{s}')
			return None
		elif clear_all:
			kj.remove_contents(target_path)
			return None

		storage_delta = [d for d in kj.storage_delta(source_storage, target_storage)]

		if not storage_delta:			# Nothing to see or do.
			return None

		if not make_changes:			# Without explicit command, show what would happen.
			for d in storage_delta:
				print(d)
			return None

		a = self.create(kj.FolderTransfer, storage_delta, target_storage.path)

		m, _ = self.select(kj.Returned, kj.Stop)
		if isinstance(m, kj.Stop):
			self.send(m, a)
			m, _ = self.select(kj.Returned)
			return kj.Aborted()

	except (OSError, ValueError) as e:
		return kj.Faulted(cannot_script, str(e))

	message = m.message
	if isinstance(message, kj.Faulted):
		return message
	return None

kj.bind(script)

# Functions supporting the
# sub-commands.
def word_i(word, i):
	'''Return the i-th element of word, if its long enough. Return element or None.'''
	if i < len(word):
		return word[i]
	return None

def home_listing(self, home_path, search, grouping=False, sub_roles=False):
	'''Load the contents of home_path and search for matching roles. Return dict[str,role] or None.'''
	home = kj.open_home(home_path, grouping=grouping, sub_roles=sub_roles)
	if home is None:
		self.complete(kj.Faulted(f'cannot list "{home_path}" (does not exist or contains unexpected/incomplete materials)'))

	# Compile all the patterns.
	machine = []
	for s in search:
		try:
			m = re.compile(s)
		except re.error as e:
			self.complete(kj.Faulted(f'cannot list "{home_path}"', str(e)))
		machine.append(m)

	# Scan for roles matching a pattern.
	if machine:
		def match(name):
			for m in machine:
				b = m.match(name)
				if b:
					return True
			return False

		home = {k: v for k, v in home.items() if match(k)}
		if not home:
			s = ', '.join(search)
			self.complete(kj.Faulted(f'cannot list "{home_path}"', f'no roles matching "{s}"'))

	return home

def home_running(self, home):
	'''Scan lock files for the given dict of roles. Return list of those that are running.'''
	running = {}
	for k, v in home.items():
		a = self.create(kj.head_lock, v.lock.path, 'head')
		self.assign(a, k)
		m, i = self.select(kj.Ready, kj.Returned)
		if isinstance(m, kj.Returned):	# Cannot lock.
			r = self.debrief()
			running[r] = m.message	# LockedOut

	return running

def world_or_clock(s, clock):
	if clock:
		t = kj.text_to_clock(s)
		d = datetime.datetime.fromtimestamp(t, tz=kj.UTC)
		return d
	return kj.text_to_world(s)

def from_last(last):
	d = datetime.datetime.now(kj.UTC)

	if last == TimeFrame.MONTH:
		f = datetime.datetime(d.year, d.month, 1, tzinfo=d.tzinfo)
	elif last == TimeFrame.WEEK:
		dow = d.weekday()
		dom = d.day - 1
		if dom >= dow:
			f = datetime.datetime(d.year, d.month, d.day - dow, tzinfo=d.tzinfo)
		elif d.month > 1:
			t = dow - dom
			r = calendar.monthrange(d.year, d.month - 1)
			f = datetime.datetime(d.year, d.month - 1, r[1] - t, tzinfo=d.tzinfo)
		else:
			t = dow - dom
			r = calendar.monthrange(d.year - 1, 12)
			f = datetime.datetime(d.year - 1, 12, r[1] - t, tzinfo=d.tzinfo)
	elif last == TimeFrame.DAY:
		f = datetime.datetime(d.year, d.month, d.day, tzinfo=d.tzinfo)
	elif last == TimeFrame.HOUR:
		f = datetime.datetime(d.year, d.month, d.day, hour=d.hour, tzinfo=d.tzinfo)
	elif last == TimeFrame.MINUTE:
		f = datetime.datetime(d.year, d.month, d.day, hour=d.hour, minute=d.minute, tzinfo=d.tzinfo)
	elif last == TimeFrame.HALF:
		t = d.minute % 30
		m = d.minute - t
		f = datetime.datetime(d.year, d.month, d.day, hour=d.hour, minute=m, tzinfo=d.tzinfo)
	elif last == TimeFrame.QUARTER:
		t = d.minute % 15
		m = d.minute - t
		f = datetime.datetime(d.year, d.month, d.day, hour=d.hour, minute=m, tzinfo=d.tzinfo)
	elif last == TimeFrame.TEN:
		t = d.minute % 10
		m = d.minute - t
		f = datetime.datetime(d.year, d.month, d.day, hour=d.hour, minute=m, tzinfo=d.tzinfo)
	elif last == TimeFrame.FIVE:
		t = d.minute % 5
		m = d.minute - t
		f = datetime.datetime(d.year, d.month, d.day, hour=d.hour, minute=m, tzinfo=d.tzinfo)
	else:
		return None
	return f

def short_delta(d):

	t = kj.span_to_text(d.total_seconds())
	i = t.find('d')
	if i != -1:
		j = t.find('h')
		if j != -1:
			return t[:j + 1]
		return t[:i + 1]
	i = t.find('h')
	if i != -1:
		j = t.find('m')
		if j != -1:
			return t[:j + 1]
		return t[:i + 1]
	# Minutes or seconds only.
	i = t.find('.')
	if i != -1:
		i += 1
		j = t.find('s')
		if j != -1:
			e = j - i
			e = min(1, e)
			return t[:i + e] + 's'
		return t[:i] + 's'

#
#
def printer(self, role, clock, begin, tail, end, count, tags):
	try:
		if begin is not None:
			reader = rl.read_log(role.logs, begin, end, count)
		else:
			reader = rl.rewind_log(role.logs, tail, end, count)

		if clock:
			for d, t in reader:
				if self.halted:
					return kj.Aborted()
				if tags is not None:
					if t[24] not in tags:
						continue
				c = d.astimezone(tz=None)		   # To localtime.
				s = c.strftime('%Y-%m-%dt%H:%M:%S') # Normal part.
				f = c.strftime('%f')[:3]			# Up to milliseconds.
				h = '%s.%s' % (s, f)
				i = t.index(' ')
				kj.output_line(h, newline=False)
				kj.output_line(t[i:], newline=False)
			return

		for d, t in reader:
			if self.halted:
				return kj.Aborted()
			if tags is not None:
				if t[24] not in tags:
					continue
			kj.output_line(t, newline=False)

	except (KeyboardInterrupt, SystemExit) as e:
		raise e
	except Exception as e:
		condition = str(e)
		fault = kj.Faulted(condition)
		return fault
	return None

kj.bind(printer)

# 2025-08-09T04:49:30.553 > <0000000f>ListenConnect - Forward Xy to <00000015> (from <0000004a>)
# 2025-08-09T04:49:30.553 < <00000015>server - Received Accepted from <0000004a>
AMPERSAND_TAG = 24
OBJECT_NAME = 36
TIME_STAMP = 23

def sampler(self, role, clock, begin, tail, end, count, sample):
	try:
		if begin is not None:
			reader = rl.read_log(role.logs, begin, end, count)
		else:
			reader = rl.rewind_log(role.logs, tail, end, count)

		for d, t in reader:
			if self.halted:
				return kj.Aborted()
			# Sample log.
			if t[AMPERSAND_TAG] != '&':
				continue
			# Begining of logged text, after name and dash.
			dash = t.find(' - ', OBJECT_NAME)
			if dash == -1:
				continue
			# Name and parens.
			text = t[dash + 3:-1]
			if text[-1] != ')':
				continue
			# Separate name from values.
			left = text.find(' (')
			if left == -1:
				continue
			# Name and values.
			name = text[:left]
			if name != sample:
				continue
			parens = text[left+2:-1]
			comma = parens.split(',')
			values = [c.split('=')[1] for c in comma]
			tabs = '\t'.join(values)
			stamp = t[0:TIME_STAMP]
			if clock:
				c = d.astimezone(tz=None)			# To localtime.
				s = c.strftime('%Y-%m-%dt%H:%M:%S') # Normal part.
				f = c.strftime('%f')[:3]			# Up to milliseconds.
				stamp = '%s.%s' % (s, f)
			else:
				stamp = t[0:TIME_STAMP]

			kj.output_line(f'{stamp}\t{tabs}')
	except (KeyboardInterrupt, SystemExit) as e:
		raise e
	except Exception as e:
		condition = str(e)
		fault = kj.Faulted(condition)
		return fault
	return None

kj.bind(sampler)

#
#
table = [
	# CRUD for a set of role definitions.
	create,
	add,
	list_,
	update,
	delete,
	destroy,

	# Orchestration of the processes, i.e. executing
	# images of the role definitions.
	run,
	start,
	stop,
	status,

	# Access to records of execution.
	log,
	history,
	returned,
	edit,

	# Network administration.
	network,
	ping,

	# Software distribution.
	resource,
	model,
	script,
]

# For package scripting.
def main():
	kj.create(cli, object_table=table, strict=False)

if __name__ == '__main__':
	main()



# test_main_return.py
import kipjak as kj

def main(self, return_an_int: bool=False, return_an_int_any: int=None):
	if return_an_int:
		return 42
	if return_an_int_any is not None:
		return (return_an_int_any, kj.Integer8())
	pass

kj.bind(main)

if __name__ == '__main__':
	kj.create(main)



#
#
import uuid
import kipjak as kj

from enum import Enum

__all__ = [
	'AutoTypes',
	'PlainTypes',
	'ContainerTypes',
	'SpecialTypes',
	'TimeTypes',
	'PointerTypes',
	'Item',
	'Brackets',
	'Letter',
	'Question',
	'Plus',
	'Parentheses',
	'Cat',
	'State',
	'MACHINE_STATE',
	'encode_decode'
]

#
#
class AutoTypes(object):
	def __init__(self, a=True, b=42, c=1.234,
			d=bytearray('Hello in bytes', 'ascii'),
			e=b'Hello in chars', f='Hello in codepoints', g=uuid.uuid4()):
		self.a = a
		self.b = b
		self.c = c
		self.d = d
		self.e = e
		self.f = f
		self.g = g

kj.bind_message(AutoTypes)

#
#
class MACHINE_STATE(Enum):
	INITIAL=1
	NEXT=2

class PlainTypes(object):
	def __init__(self, a=None, b=None, c=None, d=None, e=None,
			f=None, g=None, h=None, i=None, j=None, k=None,
			l=None, m=None, n=None, o=None, p=None, q=None):
		self.a = a or True
		self.b = b or int()
		self.c = c or bytes()
		self.d = d or str()
		self.e = e or int()
		self.f = f or int()
		self.g = g or int()
		self.h = h or int()
		self.i = i or int()
		self.j = j or int()
		self.k = k or float()
		self.l = l or float()
		self.m = m or bytearray()
		self.n = n or bytes()
		self.o = o or str()
		self.p = p or MACHINE_STATE.INITIAL
		self.q = q or uuid.uuid4()

kj.bind_message(PlainTypes,
	a=kj.Boolean(),
	b=kj.Byte(),
	c=kj.Character(),
	d=kj.Rune(),
	e=kj.Integer2(),
	f=kj.Integer4(),
	g=kj.Integer8(),
	h=kj.Unsigned2(),
	i=kj.Unsigned4(),
	j=kj.Unsigned8(),
	k=kj.Float4(),
	l=kj.Float8(),
	m=kj.Block(),
	n=kj.String(),
	o=kj.Unicode(),
	p=kj.Enumeration(MACHINE_STATE),
	q=kj.UUID(),
)

#
#
class ContainerTypes(object):
	def __init__(self, a=None, b=None, c=None, d=None, e=None,
			f=None, g=None, h=None, i=None, j=None, k=None, l=None):
		self.a = a or [bytes()] * 8
		self.b = b or [1.0, 1.1, 1.2]
		self.c = c or set([1, 2, 4, 8, 16])
		self.d = d or {"left": "right", "up": "down", "in": "out"}
		self.e = e or kj.deque([4, 5, 6])
		self.f = f or PlainTypes()
		self.g = g or [[bytes()] * 4] * 4
		self.h = h or [[2.0, 2.1, 2.2], [3.0, 3.1, 3.2], [4.0, 4.1, 4.2]]
		self.i = i or {42: {"LEFT": "RIGHT", "UP": "DOWN"}}
		self.j = j or kj.deque([kj.deque([14, 15, 16]), kj.deque([]), kj.deque([24, 25, 26])])
		self.k = k or [[bytes()]] * 4
		self.l = l or [[AutoTypes()] * 2]

kj.bind_message(ContainerTypes,
	a=kj.ArrayOf(kj.String, 8),
	b=kj.VectorOf(kj.Float8),
	c=kj.SetOf(kj.Integer8),
	d=kj.MapOf(kj.Unicode, kj.Unicode),
	e=kj.DequeOf(kj.Integer2),
	f=kj.UserDefined(PlainTypes),

	g=kj.ArrayOf(kj.ArrayOf(kj.String, 4), 4),
	h=kj.VectorOf(kj.VectorOf(kj.Float8)),
	i=kj.MapOf(kj.Integer8, kj.MapOf(kj.Unicode, kj.Unicode)),
	j=kj.DequeOf(kj.DequeOf(kj.Integer2)),
	k=kj.ArrayOf(kj.VectorOf(kj.String), 4),
	l=kj.VectorOf(kj.ArrayOf(kj.UserDefined(AutoTypes), 2)),
)

#
#
auto_types = kj.def_type(AutoTypes)
plain_types = kj.def_type(PlainTypes)

class SpecialTypes(object):
	def __init__(self, a=None, b=None, c=None, d=None, e=None, f=None):
		self.a = a or PlainTypes
		self.b = b or (3, 5, 7)	 # Will lose the 7.
		self.c = c or (2, 4, 6)	 # Add the return_proxy
		self.d = d
		self.e = {'auto': AutoTypes, 'plain': PlainTypes}
		self.f = f or []

kj.bind_message(SpecialTypes,
	a=kj.Type(),
	b=kj.TargetAddress(),
	c=kj.Address(),
	d=kj.Any(),
	e=kj.MapOf(kj.Unicode(), kj.Type()),
	f=kj.VectorOf(kj.Type()),
)

class TimeTypes(object):
	def __init__(self, a=None, b=None, c=None, d=None):
		self.a = a or kj.text_to_clock('2021-07-01T03:02:01.0')
		self.b = b or kj.text_to_span('1d2h3m4.5s')
		self.c = c or kj.text_to_world('2021-07-01T03:02:01.0+01:00')
		self.d = d or kj.text_to_delta('7:00:00:00')

kj.bind_message(TimeTypes,
	a=kj.ClockTime(),
	b=kj.TimeSpan(),
	c=kj.WorldTime(),
	d=kj.TimeDelta(),
)

#
#
class Item(object):
	def __init__(self, tag=None, next=None):
		self.tag = tag or '<blank>'
		self.next = next

kj.bind_message(Item,
	tag=kj.Unicode,
	next=kj.PointerTo(Item),
)

#
#
def linked(*tag):
	p = None
	for s in reversed(tag):
		i = Item(s, p)
		p = i
	return p

def circle(*tag):
	p = None
	t = None
	for s in reversed(tag):
		i = Item(s, p)
		if p is None:
			t = i
		p = i
	if p is not None:
		t.next = p
	return p

# Different elements of an abstract syntax
# tree.
class Brackets(object):
	def __init__(self, range=None):
		self.range = range

kj.bind_message(Brackets, range=kj.Unicode)

class Cat(object):
	def __init__(self, left=None, right=None):
		self.left = left
		self.right = right

kj.bind_message(Cat,
	left=kj.PointerTo(kj.Any),
	right=kj.PointerTo(kj.Any),
)

class Letter(object):
	def __init__(self, lone=None):
		self.lone = lone

kj.bind_message(Letter, lone=kj.Unicode)

class Question(object):
	def __init__(self, optional=None):
		self.optional = optional

kj.bind_message(Question, optional=kj.PointerTo(kj.Any))

class Plus(object):
	def __init__(self, one_or_more=None):
		self.one_or_more = one_or_more

kj.bind_message(Plus, one_or_more=kj.PointerTo(kj.Any))

class Parentheses(object):
	def __init__(self, expression=None):
		self.expression = expression

kj.bind_message(Parentheses, expression=kj.PointerTo(kj.Any))

# Parse of the following regular expression;
# [0-9]+(\.[0-9]+)?
tree = Cat(Plus(Brackets('0123456789')), Question(Cat(Letter('.'), Plus(Brackets('0123456789')))))

# Graph representation of the above AST, i.e. contains the information
# that can be used to generate transition tables.
SERIAL_ID = 1

class State(object):
	def __init__(self, number=None, edge=None):
		global SERIAL_ID
		if number is None:
			number = SERIAL_ID
			SERIAL_ID += 1
		self.number = number
		self.edge = edge or {}

kj.bind_message(State,
	number=kj.Integer8,
	edge=kj.MapOf(kj.Unicode, kj.PointerTo(State)),
)

# Build the state/edge representation for the AST above, i.e. a
# graph with pointers that can refer to any state in the network.
# In this case that includes edges referring to self and edges
# that take a shortcut to the end. Note that for simplicity of
# this example, only the edges for the 0 digit are created.
accept = State()

fraction = State(edge={'0': State(edge={None: accept})})
fraction.edge['0'].edge['0'] = fraction.edge['0']

digits = State(edge={'0': State(edge={None: accept})})
digits.edge['0'].edge['0'] = digits.edge['0']
digits.edge['0'].edge['.'] = fraction

graph = digits
#
#
class PointerTypes(object):
	def __init__(self, a=None, b=None, c=None,
			d=None, e=None, f=None, g=None):
		self.a = a or bool()
		self.b = b or self.a
		self.c = c or PlainTypes()
		self.d = d or linked('a', 'b', 'c', 'd', 'e')
		self.e = e or circle('X', 'Y', 'Z')
		self.f = f or tree
		self.g = g or graph

kj.bind_message(PointerTypes,
	a=kj.PointerTo(kj.Boolean),
	b=kj.PointerTo(kj.Boolean),
	c=kj.PointerTo(kj.UserDefined(PlainTypes)),
	d=kj.PointerTo(Item),	  # Linked using next
	e=kj.PointerTo(Item),	  # Linked using next and looped to start
	f=kj.PointerTo(kj.Any),	# AST for regular expression
	g=kj.PointerTo(State),	 # Graph for state machine
)

#
#
def encode_decode(c, test):
		t = kj.UserDefined(test)
		r = kj.make(t)

		try:
			s = c.encode(r, t)
		except kj.CodecError as e:
			print(e.note)
			return False

		# Recover the application data from the given
		# shipment.
		try:
			b = c.decode(s, t)
		except kj.CodecError as e:
			print(e.note)
			return False

		return kj.equal_to(b, r)



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Define the standard messages within the async runtime.

.

.. autoclass:: Start
.. autoclass:: Returned
.. autoclass:: Stop
"""

import re
from .virtual_memory import *
from .message_memory import *
from .convert_signature import *
from .convert_type import *

__docformat__ = 'restructuredtext'

__all__ = [
	'Start',
	'Returned',
	'Stop',
	'Pause',
	'Resume',
	'Ready',
	'NotReady',
	'Ping',
	'Enquiry',
	'Discover',
	'Inspect',
	'Ack',
	'Nak',
	'Anything',
	'Faulted',
	'Aborted',
	'TimedOut',
	'TemporarilyUnavailable',
	'Busy',
	'Overloaded',
	'OutOfService',
	'ToBeConfirmed',
	'SelectTable',
	'select_list',
	'select_list_adhoc',
	'ReForm',
]

#
#
class Start(object):
	"""Notification sent to new object, from parent."""
	pass

class Returned(object):
	"""
	Notification sent to parent, from terminating object.

	:type message: message returned to the parent
	:type message: :ref:`message<kj-message>`
	"""
	def __init__(self, message: Any=None):
		self.message = message

	def cast_back(self):
		m, p = cast_back(self.message)
		return m, p

bind_message(Start)
bind_message(Returned)

#
#
class Stop(object):
	"""Initiate termination in the receiving object."""
	pass

class Pause(object):
	"""Suspend operation in the receiving object."""
	pass

class Resume(object):
	"""Resume operation in the receiving object."""
	pass

bind_message(Stop)
bind_message(Pause)
bind_message(Resume)

#
#
class Ready(object):
	"""Report a state of readiness."""
	pass

class NotReady(object):
	"""Report that currently not ready."""
	pass

bind_message(Ready, copy_before_sending=False)
bind_message(NotReady, copy_before_sending=False)

#
#
class Ping(object):
	"""Test for reachability and presence."""
	pass

class Enquiry(object):
	"""Prompt an action from receiver."""
	pass

class Discover(object):
	"""Request for capabilities."""
	pass

class Inspect(object):
	"""Request for operational status."""
	pass

class Ack(object):
	"""Report in the positive."""
	pass

class Nak(object):
	"""Report in the negative."""
	pass

bind_message(Ping, copy_before_sending=False)
bind_message(Enquiry, copy_before_sending=False)
bind_message(Discover, copy_before_sending=False)
bind_message(Inspect, copy_before_sending=False)
bind_message(Ack, copy_before_sending=False)
bind_message(Nak, copy_before_sending=False)

#
class Anything(object):
	def __init__(self, thing=None):
		self.thing = thing

bind_message(Anything, copy_before_sending=False,
	thing=Any(),
)

#
#
class Faulted(object):
	"""
	Generic error signal to interested party.

	:param condition: description of fault
	:param explanation: description of cause
	:param error_code: internal error code
	:param exit_status: recommended exit status
	"""
	def __init__(self, condition: str=None, explanation: str=None, error_code: int=None, exit_status: int=None):
		self.condition = condition or 'fault'
		self.explanation = explanation
		self.error_code = error_code
		self.exit_status = exit_status

	def __str__(self):
		if self.explanation:
			return f'{self.condition} ({self.explanation})'
		return self.condition

bind_message(Faulted)

class Aborted(Faulted):
	"""
	Asynchronous object received a :class:`~.Stop`. Terminating.

	Derived from :class:`~.Faulted`.
	"""
	def __init__(self):
		Faulted.__init__(self, 'aborted', 'user or software interrupt')

class TimedOut(Faulted):
	"""
	Asynchronous object received a timer message.

	Derived from :class:`~.Faulted`.

	:param timer: timer that was exceeded
	:type timer: :ref:`object type<kj-object-type>`
	"""
	def __init__(self, timer=None):
		if timer and hasattr(timer, '__art__'):
			t = timer.__art__.name
		else:
			t = 'ding'
		self.timer = timer
		Faulted.__init__(self, 'timed out', f'"{t}" exceeded')

class TemporarilyUnavailable(Faulted):
	"""
	Temporarily unable to deliver normal service.

	Derived from :class:`~.Faulted`.

	:param text: description of reason
	:param unavailable: names of unavailable services
	"""
	def __init__(self, text: str=None, unavailable: list[str]=None):
		Faulted.__init__(self, text)
		self.unavailable = unavailable or []

class Busy(Faulted):
	"""
	Experiencing heavy load, recovery expected. Request is rejected.

	Derived from :class:`~.Faulted`.

	:param condition: description of load
	:param explanation: description of cause
	"""
	def __init__(self, condition: str=None, explanation: str=None):
		Faulted.__init__(self, condition, explanation)

class Overloaded(Faulted):
	"""
	Experiencing heavy load, recovery unknown. Request is rejected.

	Derived from :class:`~.Faulted`.

	:param text: description of condition
	"""
	def __init__(self, text: str=None):
		Faulted.__init__(self, text)

class OutOfService(Faulted):
	"""
	Service not available until further notice. Request is rejected.

	Derived from :class:`~.Faulted`.

	:param text: description of load
	"""
	def __init__(self, text: str=None):
		Faulted.__init__(self, text)

class ToBeConfirmed(Faulted):
	"""
	An exchange is incomplete. An update is expected.

	Derived from :class:`~.Faulted`.

	:param text: description of the delay
	"""
	def __init__(self, text: str=None):
		Faulted.__init__(self, text)

bind_message(Aborted,
	condition=Unicode(),
	explanation=Unicode(),
	error_code=Integer8(),
	exit_status=Integer8(),
)

bind_message(TimedOut,
	condition=Unicode(),
	explanation=Unicode(),
	error_code=Integer8(),
	exit_status=Integer8(),
	timer=Any(),
)

bind_message(TemporarilyUnavailable,
	condition=Unicode(),
	explanation=Unicode(),
	error_code=Integer8(),
	exit_status=Integer8(),
	unavailable=VectorOf(Unicode()),
)

bind_message(Busy,
	condition=Unicode(),
	explanation=Unicode(),
	error_code=Integer8(),
	exit_status=Integer8(),
)

bind_message(Overloaded,
	condition=Unicode(),
	explanation=Unicode(),
	error_code=Integer8(),
	exit_status=Integer8(),
	request=Unicode(),
)

bind_message(OutOfService,
	condition=Unicode(),
	explanation=Unicode(),
	error_code=Integer8(),
	exit_status=Integer8(),
	request=Unicode(),
)

bind_message(ToBeConfirmed,
	condition=Unicode(),
	explanation=Unicode(),
	error_code=Integer8(),
	exit_status=Integer8(),
	request=Unicode(),
)

#
unknown = portable_to_signature(UserDefined(Unknown))

class SelectTable(object):
	def __init__(self, unique, messaging):
		self.unique = unique
		self.messaging = messaging

	def find(self, message):
		m, p, a = un_cast(message)
		s = portable_to_signature(p)
		f = self.unique.get(s, None)		# Explicit match.
		if f:
			return f[0], m, f[1]

		if a:
			for c, f in self.messaging.items():
				if isinstance(m, c):			# Base-derived match.
					return f[0], m, f[1]

		f = self.unique.get(unknown, None)	# Catch-all.
		if f:
			return f[0], m, p
		return None

def select_list(*selection):
	"""Compile the list of types into an object suitable for :meth:`~.Point.select`.

	Prepare a lookup table for efficient matching of received messages. This is used
	to create global materials during loading.

	:param selection: message types to be included
	"""
	unique = {}
	messaging = {}
	for i, t in enumerate(selection):
		p = install_type(t)
		s = portable_to_signature(p)
		unique[s] = (i, p)
		if isinstance(p, UserDefined):
			messaging[p.element] = (i, p)
	return SelectTable(unique, messaging)

def select_list_adhoc(*selection):
	'''.'''
	unique = {}
	messaging = {}
	for i, t in enumerate(selection):
		p = lookup_type(t)
		s = portable_to_signature(p)
		unique[s] = (i, p)
		if isinstance(p, UserDefined):
			messaging[p.element] = (i, p)
	return SelectTable(unique, messaging)


#
#
class ReForm(object):
	"""Text forms based on re's embedded in character decoration.

	Form is expected as plain text with "{name}" fields. The
	fields are replaced with matching values from the ``entry`` dict;

	e.g. ``'/{resource}(/{identity})?', resource='\\w+'...``

	The :meth:`~.ReForm.compile_form` method does all the prep. Returns a 2-tuple of
	compiled pattern and a list of the field names that can be
	extracted from a match.

	:param form: the text form
	:param entry: dict of named re's
	"""
	def __init__(self, form: str=None, **entry):
		self.form = form
		self.entry = entry

	def entry_names(self):
		s = [s for s in self.entry.keys()]
		s.sort()
		return s

	def compile_form(self):
		"""Prepare the materials needed for runtime matching of the form.

		Field substitution is performed and the resulting regular expression
		is compiled. The compiled expression and the list of field names
		expected in a match are returned as a 2-tuple.

		:rtype: tuple
		"""
		form = self.form
		if form is None:
			return None, None

		entry = self.entry
		sr = {k: f'(?P<{k}>{v})' for k, v in entry.items()}
		formatted = form.format(**sr)
		search = re.compile(formatted)

		return search, self.entry_names()

bind_message(ReForm, entry=MapOf(Unicode(),Unicode()))



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Engine room of the type system, the confluence of Python, Portable and
Python hints.

Type information is gathered from class/func declarations in the form of
hints. There is also an opportunity to declare type info at the point of
registration (kipjak.bind). This is all combined/unified into a single
instance of Portable, stored in the SIGNATURE_TABLE using the generated
signature as the key. THe core idea is that only one instance of any given
type (i.e. a construction of Portables) is used within the async app.
"""

import typing
import types

__all__ = [
	'convert_hint',
	'install_hint',
	'lookup_hint',
	'convert_portable',
	'install_portable',
	'lookup_portable',
	'lookup_signature',
	'install_type',
	'lookup_type',
	'install_hints',
	'def_type',
	'cast_to',
	'un_cast',
	'cast_back',
	'bool_type',
	'int_type',
	'float_type',
	'str_type',
	'bytes_type',
	'bytearray_type',
	'datetime_type',
	'timedelta_type',
	'uuid_type',
	'message_to_tag',
]

from .virtual_memory import *
from .virtual_runtime import *
from .convert_signature import *
from collections import deque
from datetime import datetime, timedelta
import uuid
from enum import Enum

# Map of signature-to-type, ensuring the single type
# instance is in use at runtime.
SIGNATURE_TABLE = {
	'boolean': Boolean(),		# Pre-loaded with the basic types.
	'integer8': Integer8(),
	'float8': Float8(),
	'byte': Byte(),
	'block': Block(),
	'character': Character(),
	'string': String(),
	'rune': Rune(),
	'unicode': Unicode(),
	'clock': ClockTime(),
	'span': TimeSpan(),
	'world': WorldTime(),
	'delta': TimeDelta(),
	'uuid': UUID(),
	'type': Type(),
	'word': Word(),
	'address': Address(),
	'target': TargetAddress(),
	'any': Any(),
}

# Direct mapping from Python hint to
# library type.
SIMPLE_TYPE = {
	bool: Boolean(),
	int: Integer8(),
	float: Float8(),
	str: Unicode(),
	bytes: String(),
	bytearray: Block(),
	datetime: WorldTime(),
	timedelta: TimeDelta(),
	uuid.UUID: UUID(),

	Boolean: Boolean(),
	Integer8: Integer8(),
	Float8: Float8(),
	Unicode: Unicode(),
	String: String(),
	Block: Block(),
	UUID: UUID(),
	WorldTime: WorldTime(),
	TimeDelta: TimeDelta(),
	ClockTime: ClockTime(),
	TimeSpan: TimeSpan(),
	Address: Address(),
	TargetAddress: TargetAddress(),
	Type: Type(),
	Word: Word(),
	Any: Any(),
}

from typing import Union, Optional, get_origin, get_args

def convert_hint(hint, then):
	"""Accept a portable, message, basic type, Python hint or None."""

	t = SIMPLE_TYPE.get(hint, None)
	if t is not None:
		return t

	try:
		if hasattr(hint, '__art__'):
			return UserDefined(hint)
		elif issubclass(hint, Enum):
			return Enumeration(hint)
	except TypeError:
		pass

	if isinstance(hint, types.GenericAlias):
		c = hint.__origin__
		a = hint.__args__
		if c == list:
			if len(a) != 1:
				raise PointConstructionError(f'expected an argument for type "{c.__name__}"')
			a0 = convert_hint(a[0], then)
			then(a0)
			return VectorOf(a0)
		elif c == dict:
			if len(a) != 2:
				raise PointConstructionError(f'expected key and value arguments for type "{c.__name__}"')
			a0 = convert_hint(a[0], then)
			a1 = convert_hint(a[1], then)
			then(a0)
			then(a1)
			return MapOf(a0, a1)
		elif c == set:
			if len(a) != 1:
				raise PointConstructionError(f'expected an argument for type "{c.__name__}"')
			a0 = convert_hint(a[0], then)
			then(a0)
			return SetOf(a0)
		elif c == deque:
			if len(a) != 1:
				raise PointConstructionError(f'expected an argument for type "{c.__name__}"')
			a0 = convert_hint(a[0], then)
			then(a0)
			return DequeOf(a0)
	elif hint == typing.Any:
		return None
	elif hint == types.NoneType:
		return None

	try:
		origin = get_origin(hint)
		if origin:
			if origin is Union:
				args = [a for a in get_args(hint)]
				return convert_hint(args[0], then)
		return None
	except (TypeError, ValueError):
		pass

	raise PointConstructionError(f'cannot convert hint "{hint}"')

def lookup(p):
	s = portable_to_signature(p)
	f = SIGNATURE_TABLE.get(s, None)
	if f is not None:
		return f

	raise PointConstructionError(f'portable "{s}" was not loaded')

def lookup_hint(t):
	"""Search the internal table for properly installed types. Return the identity type."""
	p = convert_hint(t, lookup)
	return lookup(p)

def install(p):
	s = portable_to_signature(p)
	f = SIGNATURE_TABLE.get(s, None)
	if f is None:
		SIGNATURE_TABLE[s] = p
		return p
	return f

def install_hint(t):
	"""Search the internal table for properly installed types."""
	p = convert_hint(t, install)
	return install(p)

CONVERT_PYTHON = {
	bool: Boolean(),
	int: Integer8(),
	float: Float8(),
	str: Unicode(),
	bytes: String(),
	bytearray: Block(),
	datetime: WorldTime(),
	timedelta: TimeDelta(),
	uuid.UUID: UUID(),
}

def convert_portable(p, then, bread=None):
	"""Walk the potential hierarchy of a type, patching as necessary. Return a clean Portable."""
	if bread is None:
		bread = {}

	if is_portable(p):
		if not is_container(p):
			return p	# No change.
		# Fall thru for structured processing.
	elif is_portable_class(p):
		if not is_container_class(p):
			return p()  # Promotion of simple type.
		raise ValueError(f'portable class "{p.__name__}" used in type information, instance required')
	elif hasattr(p, '__art__'):
		if issubclass(p, Enum):
			return Enumeration(p)
		return UserDefined(p)
	else:
		# Is it one of the mapped Python classes.
		try:
			e = CONVERT_PYTHON[p]
			return e
		except KeyError:
			pass
		except TypeError:   # Unhashable - list.
			pass
		raise ValueError(f'not a portable type ({p})')

	# We have an instance of structuring.
	name = p.__class__.__name__
	if isinstance(p, ArrayOf):
		p.element = convert_portable(p.element, then, bread)
		then(p.element)
	elif isinstance(p, VectorOf):
		p.element = convert_portable(p.element, then, bread)
		then(p.element)
	elif isinstance(p, SetOf):
		p.element = convert_portable(p.element, then, bread)
		then(p.element)
	elif isinstance(p, MapOf):
		p.key = convert_portable(p.key, then, bread)
		p.value = convert_portable(p.value, then, bread)
		then(p.key)
		then(p.value)
	elif isinstance(p, DequeOf):
		p.element = convert_portable(p.element, then, bread)
		then(p.element)
	elif isinstance(p, UserDefined):
		if p.element is None or not hasattr(p.element, '__art__'):
			raise ValueError(f'"{name}" is not an installed message')
	elif isinstance(p, Enumeration):
		if not issubclass(p.element, Enum):
			raise ValueError(f'"{name}" is not an enum class')
	elif isinstance(p, PointerTo):
		try:
			e = bread[id(p)]
		except KeyError:
			e = convert_portable(p.element, then, bread)
			bread[id(p)] = e
		p.element = e
		then(p.element)
	else:
		raise ValueError('unexpected container type')
	return p

def lookup_portable(t):
	"""Search the internal table for properly installed types. Return the identity type."""
	p = convert_portable(t, lookup)
	return lookup(p)

def install_portable(t):
	"""Search the internal table for properly installed types."""
	p = convert_portable(t, install)
	return install(p)

def lookup_signature(s):
	f = SIGNATURE_TABLE.get(s, None)
	return f

def install_hints(hints):
	'''Convert standard Python hints into a managed set of Portable objects. Return a 2-tuple of dict and Portable.'''
	named_type = {}
	return_type = None
	for k, v in hints.items():
		m = install_hint(v)
		if m is None:
			continue
		if k == 'return':
			return_type = m
			continue
		named_type[k] = m
	return named_type, return_type

def lookup_type(t):
	"""Search the internal table for properly installed types. Return the identity type."""
	if isinstance(t, Portable):
		return lookup_portable(t)
	return lookup_hint(t)

def install_type(t):
	"""Search the internal table for properly installed types."""
	if isinstance(t, Portable):
		return install_portable(t)
	return install_hint(t)

def def_type(hint_or_portable):
	"""
	Register an application type. Return :class:`~.Portable`.

	Register a type for use within an application. Where necessary, convert Python
	hints into :class:`~.Portable` equivalents.

	:param hint_or_portable: a type description
	:type hint_or_portable: :ref:`tip<type-reference>`
	"""
	p = install_type(hint_or_portable)
	return p

def cast_to(value, p: Portable):
	"""
	Accept Python value and type. Return a :ref:`message<kj-message>`.

	For registered messages, this is a no-op. For constructed types it's
	required for encoding/decoding to function properly.

	:param value: the application value
	:param p: portable representation of type
	"""
	if isinstance(p, UserDefined):
		return value
	return (value, p)

def un_cast(message: Any):
	"""
	Accept the results of a :func:`~.cast_to` or network receive. Returns a 3-tuple.

	Splits the :ref:`message<kj-message>` into an application value, portable
	representation and the technical details registered at runtime.

	:param message: the portable data
	"""
	if message is None:
		return None, Any(), None
	art = getattr(message, '__art__', None)
	if art:
		value = message
		p = lookup_signature(art.path)
	elif isinstance(message, tuple) and len(message) == 2 and isinstance(message[1], Portable):
		value = message[0]
		p = message[1]
	else:
		raise ValueError(f'cannot unroll {message}')
	return value, p, art

def cast_back(message: Any):
	"""
	Accept the results of a :func:`~.cast_to` or network receive. Returns a 2-tuple.

	Splits the :ref:`message<kj-message>` into an application value and a portable type.

	:param message: class instance or constructed
	"""
	if message is None:
		return None, Any()
	art = getattr(message, '__art__', None)
	if art:
		value = message
		p = lookup_signature(art.path)
	elif isinstance(message, tuple) and len(message) == 2 and isinstance(message[1], Portable):
		value = message[0]
		p = message[1]
	else:
		raise ValueError(f'cannot unroll {message}')
	return value, p

bool_type = def_type(Boolean())
int_type = def_type(Integer8())
float_type = def_type(Float8())
str_type = def_type(Unicode())
bytes_type = def_type(String())
bytearray_type = def_type(Block())
datetime_type = def_type(WorldTime())
timedelta_type = def_type(TimeDelta())
uuid_type = def_type(UUID())

def message_to_tag(message):
	if message is None:
		return '<none>'

	art = getattr(message, '__art__', None)
	if art:
		t = art.path
	elif isinstance(message, tuple) and len(message) == 2 and isinstance(message[1], Portable):
		t = portable_to_tag(message[1])
	else:
		t = '<not-an-any>'

	return t



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Management of the async runtime.

Ensure that the support for async operation is in place when the process
needs it. Ensure that support is cleared out during process termination.
"""
__docformat__ = 'restructuredtext'

import os
import sys
import time
import signal
import uuid
import datetime
import tempfile
import json
from os.path import join
from .virtual_memory import *
from .convert_memory import *
from .message_memory import *
from .convert_type import *
from .convert_type import SIGNATURE_TABLE
from .virtual_codec import *
from .json_codec import *
from .noop_codec import *
from .file_object import *
from .folder_object import *
from .virtual_runtime import *
from .virtual_point import *
from .routine_point import *
from .point_runtime import *
from .general_purpose import *
from .command_line import *
from .command_startup import *
from .object_runtime import *
from .object_logs import *
from .rolling_log import *
from .platform_system import *
from .head_lock import *
from .home_role import *
from .bind_type import *
from .object_directory import *
from .process_directory import *

__all__ = [
	'FAULTY_EXIT',
	'CommandResponse',
	'StartStop',
	'HomeRole',
	'open_role',
	'open_home',
	'create',
]

FAULTY_EXIT = 71

#
class Unassigned(object): pass

class Incomplete(Exception):
	def __init__(self, message=Unassigned):
		self.message = message

class CommandResponse(object):
	def __init__(self, command: str=None, note: str=None):
		self.command = command
		self.note = note

bind_message(CommandResponse)

class HomeFile(object):
	def __init__(self, path_name, t):
		self.file = File(path_name, t, create_default=True)
		self.value = None

	def __call__(self):
		return self.value

	def update(self, value=Unassigned):
		if value != Unassigned:
			self.value = value
		self.file.store(self.value)

	def resume(self):
		self.value = self.file.recover()
		return self.value

class NoHomeFile(object):
	def __init__(self, value=None):
		self.value = value

	def __call__(self):
		return self.value

	def update(self, value=Unassigned):
		if value != Unassigned:
			self.value = value

	def resume(self):
		return self.value

class TmpFolder(object):
	def __init__(self, path_name, t):
		self.folder = Folder(path_name)

	def __call__(self):
		return None

	def update(self, value=Unassigned):
		pass

	def resume(self):
		remove_contents(self.folder.path)

class StartStop(object):
	def __init__(self, start: datetime.datetime=None, stop: datetime.datetime=None, returned: Any=None):
		self.start = start
		self.stop = stop
		self.returned = returned

bind_message(StartStop)

START_STOPS = 16

#
class HomeRole(object):
	def __init__(self,
		unique_id=None, start_stop=None,
		settings=None,
		resource=None,
		logs=None,
		model=None, tmp=None,
		lock=None,
		log_storage=None,
		executable_file=None):
		self.unique_id = unique_id
		self.start_stop = start_stop
		self.settings = settings
		self.resource = resource
		self.logs = logs
		self.model = model
		self.tmp = tmp
		self.lock = lock
		self.log_storage = log_storage
		self.executable_file = executable_file

	def starting(self):
		s = self.start_stop()
		s.append(StartStop(start=world_now()))
		while len(s) > START_STOPS:
			s.popleft()
		self.start_stop.update()

	def stopped(self, value):
		s = self.start_stop()
		if len(s) < 1:
			return
		s = s[-1]
		s.stop = world_now()
		s.returned = value
		self.start_stop.update()

DEFAULT_STORAGE = 1024 * 1024 * 256

#
def link_resource(role, resource_path):
	executable_file = role.executable_file()
	if not executable_file:
		return
	s = os.path.split(executable_file)
	file = s[1]

	executable_resource = os.path.join(resource_path, file)
	value = Folder(executable_resource)
	setattr(role, 'resource', value)

def open_role(home_role, resource_path):
	'''Load all the details of a role from the specified location. Returns HomeRole.'''
	try:
		# Get list of names in role, less any extent.
		split = os.path.splitext
		listing = [split(s)[0] for s in os.listdir(home_role)]
	except FileNotFoundError:
		return None

	role = HomeRole()

	# Check name is in role and folder. Only then is the
	# property created and existing value loaded.
	def resume(name, t):
		if name not in listing:
			raise Incomplete(Faulted(f'cannot open "{home_role}"', f'missing property "{name}"'))
		if not hasattr(role, name):
			raise Incomplete(Faulted(f'cannot open "{home_role}"', f'unknown property "{name}"'))

		path_name = join(home_role, name)
		f = HomeFile(path_name, t)
		f.resume()
		setattr(role, name, f)

	# Create runtime object for each named property.
	resume('unique_id', UUID())
	resume('start_stop', DequeOf(UserDefined(StartStop)))
	resume('settings', MapOf(Unicode(),Any()))
	resume('log_storage', Integer8())
	resume('executable_file', Unicode())

	# Create a runtime object for potential folders.
	def link(name):
		path_name = os.path.join(home_role, name)
		value = Folder(path_name)
		setattr(role, name, value)

	link('model')
	link('tmp')
	link('logs')
	link('lock')

	link_resource(role, resource_path)
	return role

def create_role(home_role, executable, resource_path):
	Folder(home_role)
	role = HomeRole()

	# Check name is in role and folder. Only then is the
	# property created and existing value loaded.
	def create(name, t, value):
		path_name = join(home_role, name)
		f = HomeFile(path_name, t)
		f.update(value)
		setattr(role, name, f)

	# Create runtime object for each named property.
	create('unique_id', UUID(), uuid.uuid4())
	create('settings', MapOf(Unicode(), Any()), {})
	create('start_stop', DequeOf(UserDefined(StartStop)), deque())
	create('log_storage', Integer8(), DEFAULT_STORAGE)
	create('executable_file', Unicode(), executable)

	# Create a runtime object for potential folders.
	def link(name):
		path_name = os.path.join(home_role, name)
		value = Folder(path_name)
		setattr(role, name, value)

	link('model')
	link('tmp')
	link('logs')
	link('lock')

	link_resource(role, resource_path)
	return role

def create_memory_role(executable):
	role = HomeRole()

	def create(name, t, value):
		f = NoHomeFile(value)
		setattr(role, name, f)

	# Create runtime object for each named property.
	create('unique_id', UUID(), uuid.uuid4())
	create('settings', MapOf(Unicode(), Any()), {})
	create('start_stop', DequeOf(UserDefined(StartStop)), deque())
	create('log_storage', Integer8(), DEFAULT_STORAGE)
	create('executable_file', Unicode(), executable)

	# Logical storage areas for transient, non-home instance.
	# Model - command line arg or current folder.
	cwd = os.getcwd()
	model_path = CL.model_path or cwd
	model_folder = Folder(model_path)

	# Lazy completion of tmp when the object calls
	# the supporting function - tmp() in home_role.py.

	# Resource - command line arg or None.
	resource_path = CL.resource_path
	if not resource_path:
		resource_folder = None
	else:
		resource_path = os.path.abspath(resource_path)
		resource_folder = Folder(resource_path)

	setattr(role, 'model', model_folder)
	setattr(role, 'tmp', None)
	setattr(role, 'resource', resource_folder)

	return role

#
def open_home(home_path, grouping=False, sub_roles=False):
	'''Load all the roles within a folder. Return a dict of HomeRoles'''

	role_path = join(home_path, 'role')
	resource_path = join(home_path, 'resource')
	try:
		def role(s):
			if '.' in s and not sub_roles:
				return False
			elif s.startswith('group') and not grouping:
				return False
			return True

		listing = {s: open_role(join(role_path, s), resource_path) for s in os.listdir(role_path) if role(s)}
	except FileNotFoundError:
		return None
	except Incomplete:
		return None

	return listing

def object_home(executable, home_role, resource_path, sticky=False):
	'''Compile the runtime, file-based context for the current process. Return HomeRole and role.'''

	if CL.origin == ProcessOrigin.SHELL:
		role = None
	else:
		role = open_role(home_role, resource_path)

	if CL.create_role:
		if role is not None:
			raise Incomplete(Faulted(f'cannot create "{home_role}"', f'already exists'))
		role = create_role(home_role, executable, resource_path)
	elif role is None:
		if CL.origin in (ProcessOrigin.START, ProcessOrigin.START_CHILD) or sticky or CL.keep_logs:
			role = create_role(home_role, executable, resource_path)
		else:
			role = create_memory_role(executable)

	# Circular dependency around log_storage. Wrap it in a
	# condition and default to fixed value.
	log_storage = role.log_storage() if role is not None else DEFAULT_STORAGE

	logs, files_in_folder = open_logs(home_role, log_storage)
	if files_in_folder:
		logs = RollingLog(role.logs.path, files_in_folder=files_in_folder)

	rolling = isinstance(logs, RollingLog)
	if sticky or rolling:
		if role is None:
			if home_role.count('.') == 0:
				raise Incomplete(Faulted(f'cannot auto-create "{home_role}"', f'sub-roles only'))
			role = create_role(home_role, executable, resource_path)

	elif role is None:
		role = create_memory_role(executable)

	if role.tmp:
		remove_contents(role.tmp.path)

	return role, logs

def daemonize():
	"""
	Do the UNIX double-fork shuffle, see Stevens' "Advanced
	Programming in the UNIX Environment" for details (ISBN 0201563177)
	http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
	"""
	try:
		pid = os.fork()
		if pid > 0:
			# exit first parent
			sys.exit(0)
	except OSError as e:
		e = str(e)
		f = Faulted(f'cannot fork to daemonize #1 ({e})')
		raise Incomplete(f)

	# decouple from parent environment
	os.chdir("/")
	os.setsid()
	os.umask(0)

	try:
		pid = os.fork()
		if pid > 0:
			# exit second parent
			sys.exit(0)
	except OSError as e:
		e = str(e)
		f = Faulted(f'cannot fork to daemonize #2 ({e})')
		raise Incomplete(f)

	# redirect standard file descriptors
	#sys.stdout.flush()
	#sys.stderr.flush()
	#si = file(self.stdin, 'r')
	#so = file(self.stdout, 'a+')
	#se = file(self.stderr, 'a+', 0)
	#os.dup2(si.fileno(), sys.stdin.fileno())
	#os.dup2(so.fileno(), sys.stdout.fileno())
	#os.dup2(se.fileno(), sys.stderr.fileno())

	# write pidfile
	#atexit.register(self.delpid)
	#pid = str(os.getpid())
	#file(self.pidfile,'w+').write("%s\n" % pid)

def open_logs(home_role, storage):
	debug_level = CL.debug_level

	if CL.origin in (ProcessOrigin.START, ProcessOrigin.START_CHILD) or CL.keep_logs:
		bytes_in_file = 120 * LINES_IN_FILE
		files_in_folder = storage / bytes_in_file
		return None, files_in_folder
	elif debug_level is None:
		logs = log_to_nowhere
	else:
		logs = select_logs(debug_level)

	return logs, None

#def object_home():
#	"""Global access to the runtime context assumed by create_object(). Returns a HomeRole."""
#	return OBJECT_HOME.home_path
#	name_counts = ['"%s" (%d)' % (k, len(v)) for k, v in pt.thread_classes.items()]

#	executable = os.path.abspath(sys.argv[0])
#	self.trace('Working folder "%s"' % (os.getcwd()))
#	self.trace('Running object "%s"' % (object_type.__art__.path,))
#	self.trace('Class threads (%d) %s' % (len(pt.thread_classes), ','.join(name_counts)))

def start_vector(self, object_type, word, args):
	a = self.create(object_type, *word, **args)

	if CL.directory_scope == ScopeOfDirectory.LIBRARY:
		pn = PublishAs(name=CL.role_name, scope=ScopeOfDirectory.PROCESS, publisher_address=a)
		self.send(pn, PD.directory)

	while True:
		m, i = self.select(Returned, Stop)

		if i == 0:
			# Do a "fake" signaling. Sidestep all the platform machinery
			# and just set a global. It does avoid any complexities
			PS.signal_received = PS.platform_kill
		elif i == 1:
			self.send(m, a)
			m, i = self.select(Returned)

		return m.message

bind_routine(start_vector)

def run_object(home, object_type, word, args, logs, locking):
	'''Start the async runtime, lock if required and make arrangements for control-c handling.'''
	message = None
	try:
		# Install signal handlers, i.e. control-c.
		ps = PS.platform_signal
		if ps is None:
			f = Faulted(f'unknown "{PS.platform_system}" ({PS.platform_release})')
			raise Incomplete(f)
		ps()

		# Start the async runtime.
		root = start_up(logs)

		# Exclusive access to disk-based resources.
		if locking or isinstance(logs, RollingLog):
			a = root.create(head_lock, home.lock.path, 'head')
			root.assign(a, 1)
			m, i = root.select(Ready, Returned)
			if isinstance(m, Returned):	# Cannot lock.
				root.debrief()
				c = Faulted(f'role {home.lock.path} is running')
				raise Incomplete(c)

		if CL.edit_role:
			message = HR.edit_role(root, home)
			return message

		# Respond to daemon context, i.e. send output and close stdout.
		#if CL.origin == ProcessOrigin.START:	# or no_output:
		#	early_return = True
		#	object_encode(CommandResponse('background-daemon'))
		#	sys.stdout.close()
		#	os.close(1)

		# Write partial record to disk.
		home.starting()

		# Create the async object. Need to wrap in another object
		# to facilitate the control-c handling.
		a = root.create(start_vector, object_type, word, args)

		# Termination of this function is
		# either by SIGINT (control-c) or assignment by object_vector.
		running = True
		while running:
			while PS.signal_received is None:
				time.sleep(0.25)
				#signal.pause()
			sr, PS.signal_received = PS.signal_received, None

			# If it was keyboard then async object needs
			# to be bumped.
			if sr == PS.platform_kill:
				running = False
			elif sr == signal.SIGINT:
				root.send(Stop(), a)
				running = False

		m, i = root.select(Returned)		# End of start_vector.
		message = m.message

	finally:
		root.abort()					# Stop the lock if present.
		while root.working():
			root.select(Returned)
			root.debrief()

	home.stopped(message)

	#if early_return:		# Already sent message. Silence any output.
	#	return None

	return message

def object_encode(value):
	'''Put the encoding of the final result, on to stdout.'''
	pretty_format = not CL.child_process
	if CL.full_output:
		codec = CodecJson(pretty_format=pretty_format)
		output = codec.encode(value, Any())
		sys.stdout.write(output)
		sys.stdout.write('\n')
		return
	codec = CodecNoop()
	js = codec.encode(value, Any())
	value = js['value'][1]
	indent = 4 if pretty_format else None
	output = json.dumps(value, indent=indent)
	sys.stdout.write(output)
	sys.stdout.write('\n')

def object_error(fault):
	'''Print the final result - an error - on the console.'''
	p = sys.argv[0]
	sys.stderr.write(f'{p}: {fault}\n')

def object_output(value):
	'''Put the final output into an output file or on stdout.'''
	if value is None:
		return
	output_file = CL.output_file

	try:
		if output_file:
			f = File(output_file, Any())
			f.store(value)
			return
		object_encode(value)
		return
	except OSError as e:
		value = Faulted(str(e))
		PB.exit_status = e.args[0]
		if not CL.full_output:
			object_error(value)
			return
	except CodecError as e:
		value = Faulted(str(e))
		PB.exit_status = FAULTY_EXIT
		if not CL.full_output:
			object_error(value)
			return

	# Single, unmanaged attempt to output a failed object
	# output, i.e. cant open output file or failed encoding.
	object_encode(value)

#
def create(object_type, object_table=None, environment_variables=None, sticky: bool=False, strict: bool=True):
	"""Creates an async process shim around a "main" async object. Returns nothing.

	:param object_type: type of object to be instantiated
	:type object_type: :ref:`object type<kj-object-type>`
	:param object_table: sub-commands accepted by this object type
	:type object_table: a list of registered functions
	:param environment_variables: container of values to be extracted from environment
	:type environment_variables: :ref:`message<kj-message>`
	:param sticky: object requires persistent storage
	:param strict: object requires all args to match
	"""
	early_return = False
	try:
		# Break down the command line with reference to the
		# name/type information in the object type.
		if object_table is None:
			executable, argument, word = command_arguments(object_type)
		else:
			executable, argument, word = command_sub_arguments(object_type, object_table, strict=strict)

		bp = breakpath(executable)
		name = bp[1]

		# Compose the location of file-based materials.
		home_path = CL.home_path or DEFAULT_HOME
		role_name = CL.role_name or name

		home_path = os.path.abspath(home_path)
		home_role = join(home_path, 'role', role_name)
		resource_path = join(home_path, 'resource')

		# Extract values from the environment with reference
		# to the name/type info in the variables object.
		command_variables(environment_variables)

		# Resume the appropriate operational context, i.e. home.
		home, logs = object_home(executable, home_role, resource_path, sticky=sticky)

		HR.home_path = home_path
		HR.role_name = role_name
		HR.home_role = home_role

		# Transfer folder info visible to object, see home_role.model().
		HR.model = home.model
		HR.tmp = home.tmp
		HR.resource = home.resource

		if CL.dump_types:
			table = sorted(SIGNATURE_TABLE.keys())
			for t in table:
				print(t)
			raise Incomplete(None)

		if CL.create_role:
			home.settings.update(argument)
			t = [a for a in argument.keys()]
			if not t:
				t = ['empty']
			c = CommandResponse('create-role', ','.join(t))
			raise Incomplete(c)

		expect_settings = CL.update_role or CL.dump_role
		expect_settings = expect_settings or CL.factory_reset or CL.delete_role
		if expect_settings and not isinstance(home.settings, HomeFile):
			raise Incomplete(Faulted(f'no settings available "{home_role}"'))

		# Non-operational features, i.e. command object not called.
		# Current arguments not included.
		if CL.factory_reset:
			home.settings.update({})
			c = CommandResponse('factory-reset')
			raise Incomplete(c)

		if CL.dump_role:
			settings = (home.settings(), MapOf(Unicode(), Any()))
			object_encode(settings)
			raise Incomplete(None)		# Settings on stdout.

		# Add/override current settings with values from
		# the command line.
		settings = home.settings()
		for k, v in argument.items():
			settings[k] = v

		if CL.update_role:
			home.settings.update()
			t = [a for a in settings.keys()]
			if not t:
				t = ['empty']
			c = CommandResponse('update-role', ','.join(t))
			raise Incomplete(c)

		if CL.delete_role:
			try:
				remove_folder(HR.home_role)
			except OSError as e:
				raise Incomplete(Faulted('Cannot delete role', str(e)))
			c = CommandResponse('delete-role')
			raise Incomplete(c)

		if CL.help:
			#command_help(object_type, argument)
			raise Incomplete(None)

		if CL.origin == ProcessOrigin.START:
			object_encode(CommandResponse('background-daemon'))
			sys.stdout.close()
			os.close(1)
			daemonize()
			early_return = True

		rolling = isinstance(logs, RollingLog)
		locking = sticky or rolling

		args = kv_arguments(settings)

		output = run_object(home, object_type, word, args, logs, locking)
	except (CodecError, ValueError, KeyError) as e:
		s = str(e)
		output = Faulted(s)
	except Incomplete as e:
		output = e.message

	if HR.temp_dir is not None:
		HR.temp_dir.cleanup()

	def ending(output):
		exit_status = 0
		if isinstance(output, Faulted):
			# Need a non-zero exit for command-line scenario but also a value
			# that ProcessObject can detect as part of framework operations,
			# i.e. and still make use of the fault, rather than discarding
			# the event.
			exit_status = output.exit_status if output.exit_status is not None else FAULTY_EXIT
			if not CL.full_output and not early_return:
				object_error(output)
				return exit_status
		if not early_return:
			object_output(output)
		return exit_status

	# Make available to tear_down (atexit) and also for checking
	# within unit tests.
	PB.output_value, PB.exit_status = output, ending(output)



# test_main_args.py
import datetime
import uuid
import kipjak as kj

from test_person import *

class Catcher(kj.Point, kj.Stateless):
	def __init__(self, ):
		kj.Point.__init__(self)
		kj.Stateless.__init__(self)

def Catcher_Start(self, message):
	pass

def Catcher_dict_str_list_Person(self, message):
	person = []
	for v in message.values():
		for p in v:
			person.append(p.given_name)

	csv = ','.join(person)
	self.console(table=csv)

def Catcher_int(self, message):
	self.console(message=message)

def Catcher_float(self, message):
	self.console(message=message)

def Catcher_Person(self, message):
	self.console(person=message.given_name)

def Catcher_datetime(self, message):
	self.console(message=message)

def Catcher_UUID(self, message):
	self.complete()

kj.bind(Catcher, dispatch=(kj.Start,
	dict[str,list[Person]],
	int, float,
	Person,
	datetime.datetime,uuid.UUID))

#
table_type = kj.def_type(dict[str,list[Person]])

class Main(kj.Point, kj.Stateless):
	def __init__(self, table: dict[str,list[Person]]=None,
		count: int=10, ratio: float=0.5,
		who: Person=None, when: datetime.datetime=None,
		unique_id: uuid.UUID=None):
		kj.Point.__init__(self)
		kj.Stateless.__init__(self)
		self.table = table or dict(recent=[Person('Felicity'), Person('Frederic')])
		self.who = who or Person('Wilfred')
		self.when = when or kj.world_now()
		self.unique_id = unique_id or uuid.uuid4()
		self.count = count
		self.ratio = ratio

def Main_Start(self, message):
	j = self.create(Catcher)
	self.send(kj.cast_to(self.table, table_type), j)
	self.send(kj.cast_to(self.count, kj.int_type), j)
	self.send(kj.cast_to(self.ratio, kj.float_type), j)
	self.send(self.who, j)
	self.send(kj.cast_to(self.when, kj.datetime_type), j)
	self.send(kj.cast_to(self.unique_id, kj.uuid_type), j)

def Main_Returned(self, message):
	# Catcher terminates on receiving UUID.
	self.complete()

def Main_Stop(self, message):
	self.complete()

kj.bind(Main, dispatch=(kj.Start, kj.Returned, kj.Stop))

if __name__ == '__main__':
	kj.create(Main)



# test_main_args.py
import uuid
import kipjak as kj

from test_person import *

__all__ = [
	'INITIAL',
	'READY',
	'Main',
	'Main_INITIAL_Start',
	'Main_INITIAL_Faulted',
	'Main_INITIAL_Unknown',
	'Main_READY_int',
	'Main_READY_list_int',
	'Main_READY_Person',
	'Main_READY_dict_UUID_Person',
	'Main_READY_TimedOut',
	'Main_READY_Faulted',
	'Main_READY_Stop',
	'Main_READY_Unknown',
]

class INITIAL: pass
class READY: pass

class Main(kj.Point, kj.StateMachine):
	def __init__(self, height: int=8, width: int=8, value: float=0.125):
		kj.Point.__init__(self)
		kj.StateMachine.__init__(self, INITIAL)
		self.height = height
		self.width = width
		self.value = value

def Main_INITIAL_Start(self, message):
	self.console(height=self.height, width=self.width, value=self.value)
	return READY

def Main_INITIAL_Faulted(self, message):
	self.console(wtf=message)
	return INITIAL

def Main_INITIAL_Unknown(self, message):
	self.console(wtf=message)
	return INITIAL

def Main_READY_int(self, message):
	self.console(message=message)
	return READY

def Main_READY_list_int(self, message):
	self.console(message=message)
	return READY

def Main_READY_Person(self, message):
	self.console(message=message)
	return READY

def Main_READY_dict_UUID_Person(self, message):
	self.console(message=message)
	return READY

def Main_READY_TimedOut(self, message):
	self.console(message=message)
	return READY

def Main_READY_Faulted(self, message):
	self.console(message=message)
	return READY

def Main_READY_Stop(self, message):
	table = [[self.value] * self.height] * self.width
	self.complete(table)

def Main_READY_Unknown(self, message):
	self.console(message=message)
	return READY

MAIN_DISPATCH = {
	INITIAL: ((kj.Start, kj.Faulted, kj.Unknown), ()),
	READY: ((int, list[int], Person, dict[uuid.UUID, Person], kj.TimedOut, kj.Faulted, kj.Stop, kj.Unknown), ()),
}

kj.bind(Main,
	MAIN_DISPATCH,
	return_type=kj.VectorOf(kj.VectorOf(float)))

if __name__ == '__main__':
	kj.create(Main)



# Author: Scott Woods <scott.suzuki@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Any general-purpose class or function that does not belong elsewhere.

"""

__docformat__ = 'restructuredtext'

import os
import sys
import random
from enum import Enum

from .virtual_memory import *
from .convert_memory import *
from .message_memory import *

__all__ = [
	'Gas',
	'breakpath',
	'output_line',
	'short_delta',
	'spread_out',
]

random.seed()


#
#
class Gas(object):
	"""Build an object from the specified key-value args.

	Create an attribute for each of the named arguments. The
	values are subsequently available using standard object
	member access.

	:param kv: map of names and value
	"""
	def __init__(self, **kv):
		for k, v in kv.items():
			setattr(self, k, v)

#
#
def breakpath(p):
	"""Break apart the full path into folder, file and extent (3-tuple)."""
	p, f = os.path.split(p)
	name, e = os.path.splitext(f)
	return p, name, e

#
#
def output_line(line, tab=0, newline=True, **kv):
	if kv:
		line = line.format(**kv)

	if tab:
		sys.stdout.write('+   ' * tab)

	sys.stdout.write(line)
	if newline:
		sys.stdout.write('\n')

#
def short_delta(d):
	t = span_to_text(d.total_seconds())
	i = t.find('d')
	if i != -1:
		j = t.find('h')
		if j != -1:
			return t[:j + 1]
		return t[:i + 1]
	i = t.find('h')
	if i != -1:
		j = t.find('m')
		if j != -1:
			return t[:j + 1]
		return t[:i + 1]

	# Minutes or seconds only.
	i = t.find('.')
	if i != -1:
		i += 1
		j = t.find('s')
		if j != -1:
			e = j - i
			e = min(1, e)
			return t[:i + e] + 's'
		return t[:i] + 's'
	return t

def spread_out(period: float, delta: int=25):
	'''
	Adjust a base value in a random way. Return a float.


	:param period: base time period
	:param delta: range of adjustment as percent denominator
	'''
	lo = 100 - delta
	hi = 100 + delta
	cf = random.randrange(lo, hi) / 100
	return period * cf



# test_echo.py
import kipjak as kj
import test_library

def main(self):
	echo = self.create(kj.ProcessObject, test_library.main)		# Load the library process.

	self.send(kj.Ack(), echo)									# Request and
	m, i = self.select(kj.Ack, kj.Faulted, kj.Stop)			# response.
	assert isinstance(m, kj.Ack)

	m, i = self.select()			# response.
	return kj.Aborted()

	# Optional housekeeping.
	# self.send(kj.Stop(), echo)									# Unload the library.
	# m, i = self.select(kj.Returned, kj.Faulted, kj.Stop)		# Gone.
	# assert isinstance(m, kj.Returned)

kj.bind(main)

if __name__ == '__main__':
	kj.create(main)
